/*
 * Copyright (c) 2017, Alliance for Open Media. All rights reserved
 *
 * This source code is subject to the terms of the BSD 2 Clause License and
 * the Alliance for Open Media Patent License 1.0. If the BSD 2 Clause License
 * was not distributed with this source code in the LICENSE file, you can
 * obtain it at www.aomedia.org/license/software. If the Alliance for Open
 * Media Patent License 1.0 was not distributed with this source code in the
 * PATENTS file, you can obtain it at www.aomedia.org/license/patent.
 */

#include "./aom_config.h"
#include "av1/common/entropy.h"

static const aom_cdf_prob
    av1_default_txb_skip_cdfs[TOKEN_CDF_Q_CTXS][TX_SIZES][TXB_SKIP_CONTEXTS]
                             [CDF_SIZE(2)] = {
                               {
                                   { { AOM_CDF2(32605) },
                                     { AOM_CDF2(10627) },
                                     { AOM_CDF2(13167) },
                                     { AOM_CDF2(22671) },
                                     { AOM_CDF2(18710) },
                                     { AOM_CDF2(26810) },
                                     { AOM_CDF2(32126) },
                                     { AOM_CDF2(3637) },
                                     { AOM_CDF2(16905) },
                                     { AOM_CDF2(28677) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) } },
                                   { { AOM_CDF2(32613) },
                                     { AOM_CDF2(12454) },
                                     { AOM_CDF2(14380) },
                                     { AOM_CDF2(25061) },
                                     { AOM_CDF2(18405) },
                                     { AOM_CDF2(27262) },
                                     { AOM_CDF2(32308) },
                                     { AOM_CDF2(2819) },
                                     { AOM_CDF2(15663) },
                                     { AOM_CDF2(28353) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) } },
                                   { { AOM_CDF2(32209) },
                                     { AOM_CDF2(10135) },
                                     { AOM_CDF2(14508) },
                                     { AOM_CDF2(23776) },
                                     { AOM_CDF2(21415) },
                                     { AOM_CDF2(25031) },
                                     { AOM_CDF2(31713) },
                                     { AOM_CDF2(5176) },
                                     { AOM_CDF2(16175) },
                                     { AOM_CDF2(29220) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) } },
                                   { { AOM_CDF2(26786) },
                                     { AOM_CDF2(11656) },
                                     { AOM_CDF2(17377) },
                                     { AOM_CDF2(26137) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(26600) },
                                     { AOM_CDF2(31359) },
                                     { AOM_CDF2(11087) },
                                     { AOM_CDF2(25919) },
                                     { AOM_CDF2(31989) },
                                     { AOM_CDF2(15766) },
                                     { AOM_CDF2(26756) },
                                     { AOM_CDF2(29848) } },
                                   { { AOM_CDF2(21611) },
                                     { AOM_CDF2(5461) },
                                     { AOM_CDF2(28087) },
                                     { AOM_CDF2(25817) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(28672) },
                                     { AOM_CDF2(32208) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) } },
                               },
                               {
                                   { { AOM_CDF2(31574) },
                                     { AOM_CDF2(10880) },
                                     { AOM_CDF2(13298) },
                                     { AOM_CDF2(20932) },
                                     { AOM_CDF2(17307) },
                                     { AOM_CDF2(25362) },
                                     { AOM_CDF2(31521) },
                                     { AOM_CDF2(4062) },
                                     { AOM_CDF2(15737) },
                                     { AOM_CDF2(27189) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) } },
                                   { { AOM_CDF2(32595) },
                                     { AOM_CDF2(11226) },
                                     { AOM_CDF2(12558) },
                                     { AOM_CDF2(22364) },
                                     { AOM_CDF2(17483) },
                                     { AOM_CDF2(26136) },
                                     { AOM_CDF2(32081) },
                                     { AOM_CDF2(3509) },
                                     { AOM_CDF2(14439) },
                                     { AOM_CDF2(26145) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) } },
                                   { { AOM_CDF2(32485) },
                                     { AOM_CDF2(12729) },
                                     { AOM_CDF2(14812) },
                                     { AOM_CDF2(22261) },
                                     { AOM_CDF2(20672) },
                                     { AOM_CDF2(25408) },
                                     { AOM_CDF2(31463) },
                                     { AOM_CDF2(3565) },
                                     { AOM_CDF2(15559) },
                                     { AOM_CDF2(27508) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) } },
                                   { { AOM_CDF2(30478) },
                                     { AOM_CDF2(10847) },
                                     { AOM_CDF2(15746) },
                                     { AOM_CDF2(25044) },
                                     { AOM_CDF2(23164) },
                                     { AOM_CDF2(25746) },
                                     { AOM_CDF2(31778) },
                                     { AOM_CDF2(7642) },
                                     { AOM_CDF2(22163) },
                                     { AOM_CDF2(30328) },
                                     { AOM_CDF2(8536) },
                                     { AOM_CDF2(24702) },
                                     { AOM_CDF2(30147) } },
                                   { { AOM_CDF2(26684) },
                                     { AOM_CDF2(3208) },
                                     { AOM_CDF2(17644) },
                                     { AOM_CDF2(27445) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(30341) },
                                     { AOM_CDF2(32556) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) },
                                     { AOM_CDF2(16384) } },
                               },
                               { { { AOM_CDF2(29941) },
                                   { AOM_CDF2(9962) },
                                   { AOM_CDF2(11511) },
                                   { AOM_CDF2(19393) },
                                   { AOM_CDF2(14863) },
                                   { AOM_CDF2(23680) },
                                   { AOM_CDF2(30957) },
                                   { AOM_CDF2(3960) },
                                   { AOM_CDF2(15120) },
                                   { AOM_CDF2(25656) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) } },
                                 { { AOM_CDF2(32541) },
                                   { AOM_CDF2(9637) },
                                   { AOM_CDF2(11428) },
                                   { AOM_CDF2(20348) },
                                   { AOM_CDF2(16737) },
                                   { AOM_CDF2(24878) },
                                   { AOM_CDF2(31599) },
                                   { AOM_CDF2(3938) },
                                   { AOM_CDF2(13268) },
                                   { AOM_CDF2(23916) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) } },
                                 { { AOM_CDF2(32560) },
                                   { AOM_CDF2(14332) },
                                   { AOM_CDF2(14899) },
                                   { AOM_CDF2(21998) },
                                   { AOM_CDF2(22306) },
                                   { AOM_CDF2(27670) },
                                   { AOM_CDF2(31776) },
                                   { AOM_CDF2(4071) },
                                   { AOM_CDF2(15952) },
                                   { AOM_CDF2(26342) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) } },
                                 { { AOM_CDF2(31747) },
                                   { AOM_CDF2(11539) },
                                   { AOM_CDF2(15576) },
                                   { AOM_CDF2(24995) },
                                   { AOM_CDF2(25295) },
                                   { AOM_CDF2(25947) },
                                   { AOM_CDF2(32154) },
                                   { AOM_CDF2(6818) },
                                   { AOM_CDF2(21562) },
                                   { AOM_CDF2(28731) },
                                   { AOM_CDF2(3421) },
                                   { AOM_CDF2(21167) },
                                   { AOM_CDF2(30760) } },
                                 { { AOM_CDF2(30272) },
                                   { AOM_CDF2(7198) },
                                   { AOM_CDF2(22483) },
                                   { AOM_CDF2(27619) },
                                   { AOM_CDF2(27483) },
                                   { AOM_CDF2(29545) },
                                   { AOM_CDF2(32353) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) } } },
                               { { { AOM_CDF2(30198) },
                                   { AOM_CDF2(7807) },
                                   { AOM_CDF2(9940) },
                                   { AOM_CDF2(19103) },
                                   { AOM_CDF2(14088) },
                                   { AOM_CDF2(22936) },
                                   { AOM_CDF2(30320) },
                                   { AOM_CDF2(3227) },
                                   { AOM_CDF2(11686) },
                                   { AOM_CDF2(21633) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) } },
                                 { { AOM_CDF2(32322) },
                                   { AOM_CDF2(6667) },
                                   { AOM_CDF2(9014) },
                                   { AOM_CDF2(18042) },
                                   { AOM_CDF2(14207) },
                                   { AOM_CDF2(22732) },
                                   { AOM_CDF2(30904) },
                                   { AOM_CDF2(3422) },
                                   { AOM_CDF2(10341) },
                                   { AOM_CDF2(19457) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) } },
                                 { { AOM_CDF2(32463) },
                                   { AOM_CDF2(10932) },
                                   { AOM_CDF2(14618) },
                                   { AOM_CDF2(21939) },
                                   { AOM_CDF2(20793) },
                                   { AOM_CDF2(26929) },
                                   { AOM_CDF2(31366) },
                                   { AOM_CDF2(4351) },
                                   { AOM_CDF2(12704) },
                                   { AOM_CDF2(22868) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) } },
                                 { { AOM_CDF2(31987) },
                                   { AOM_CDF2(9885) },
                                   { AOM_CDF2(21616) },
                                   { AOM_CDF2(23702) },
                                   { AOM_CDF2(23316) },
                                   { AOM_CDF2(28245) },
                                   { AOM_CDF2(32112) },
                                   { AOM_CDF2(5654) },
                                   { AOM_CDF2(18625) },
                                   { AOM_CDF2(25878) },
                                   { AOM_CDF2(1762) },
                                   { AOM_CDF2(14824) },
                                   { AOM_CDF2(26877) } },
                                 { { AOM_CDF2(32374) },
                                   { AOM_CDF2(14704) },
                                   { AOM_CDF2(24688) },
                                   { AOM_CDF2(28913) },
                                   { AOM_CDF2(29407) },
                                   { AOM_CDF2(30985) },
                                   { AOM_CDF2(32453) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) },
                                   { AOM_CDF2(16384) } } },
                             };

static const aom_cdf_prob
    av1_default_dc_sign_cdfs[TOKEN_CDF_Q_CTXS][PLANE_TYPES][DC_SIGN_CONTEXTS]
                            [CDF_SIZE(2)] = {
                              { {
                                    { AOM_CDF2(128 * 125) },
                                    { AOM_CDF2(128 * 102) },
                                    { AOM_CDF2(128 * 147) },
                                },
                                {
                                    { AOM_CDF2(128 * 119) },
                                    { AOM_CDF2(128 * 101) },
                                    { AOM_CDF2(128 * 135) },
                                } },
                              { {
                                    { AOM_CDF2(128 * 125) },
                                    { AOM_CDF2(128 * 102) },
                                    { AOM_CDF2(128 * 147) },
                                },
                                {
                                    { AOM_CDF2(128 * 119) },
                                    { AOM_CDF2(128 * 101) },
                                    { AOM_CDF2(128 * 135) },
                                } },
                              { {
                                    { AOM_CDF2(128 * 125) },
                                    { AOM_CDF2(128 * 102) },
                                    { AOM_CDF2(128 * 147) },
                                },
                                {
                                    { AOM_CDF2(128 * 119) },
                                    { AOM_CDF2(128 * 101) },
                                    { AOM_CDF2(128 * 135) },
                                } },
                              { {
                                    { AOM_CDF2(128 * 125) },
                                    { AOM_CDF2(128 * 102) },
                                    { AOM_CDF2(128 * 147) },
                                },
                                {
                                    { AOM_CDF2(128 * 119) },
                                    { AOM_CDF2(128 * 101) },
                                    { AOM_CDF2(128 * 135) },
                                } },
                            };

static const aom_cdf_prob
    av1_default_eob_multi16_cdfs[TOKEN_CDF_Q_CTXS][PLANE_TYPES][2][CDF_SIZE(
        5)] = {
      { { { AOM_CDF5(1063, 1368, 2747, 5584) },
          { AOM_CDF5(269, 442, 1012, 4118) } },
        { { AOM_CDF5(5514, 7645, 11953, 19700) },
          { AOM_CDF5(3979, 6565, 11760, 20491) } } },
      { { { AOM_CDF5(2689, 3569, 6487, 11023) },
          { AOM_CDF5(469, 787, 1694, 6678) } },
        { { AOM_CDF5(9194, 11895, 17248, 24761) },
          { AOM_CDF5(6407, 10248, 16706, 25074) } } },
      { { { AOM_CDF5(4637, 6257, 10483, 16764) },
          { AOM_CDF5(686, 1307, 2637, 9683) } },
        { { AOM_CDF5(12745, 15999, 21878, 28227) },
          { AOM_CDF5(9092, 14278, 20810, 28556) } } },
      { { { AOM_CDF5(6544, 9316, 15031, 22497) },
          { AOM_CDF5(1059, 2140, 4290, 12545) } },
        { { AOM_CDF5(20492, 23668, 28204, 31451) },
          { AOM_CDF5(14289, 21097, 25960, 31272) } } },
    };

static const aom_cdf_prob
    av1_default_eob_multi32_cdfs[TOKEN_CDF_Q_CTXS][PLANE_TYPES][2][CDF_SIZE(
        6)] = {
      { { { AOM_CDF6(722, 898, 1491, 2584, 6481) },
          { AOM_CDF6(199, 308, 781, 2096, 6618) } },
        { { AOM_CDF6(6788, 8626, 12634, 17791, 26086) },
          { AOM_CDF6(4005, 6267, 10129, 15919, 29241) } } },
      { { { AOM_CDF6(1699, 2138, 3410, 5506, 11583) },
          { AOM_CDF6(261, 436, 1430, 3761, 10049) } },
        { { AOM_CDF6(11733, 13930, 18373, 23020, 29777) },
          { AOM_CDF6(7461, 12117, 19218, 25087, 31599) } } },
      { { { AOM_CDF6(2969, 3841, 5727, 9044, 16543) },
          { AOM_CDF6(407, 736, 1930, 5443, 13281) } },
        { { AOM_CDF6(15519, 18234, 22613, 26662, 31670) },
          { AOM_CDF6(9677, 15468, 23735, 29377, 32496) } } },
      { { { AOM_CDF6(4072, 5629, 8217, 12353, 21017) },
          { AOM_CDF6(755, 1245, 3058, 7733, 17443) } },
        { { AOM_CDF6(24075, 26423, 29061, 30983, 32407) },
          { AOM_CDF6(14789, 20299, 30158, 31898, 32478) } } },
    };

static const aom_cdf_prob
    av1_default_eob_multi64_cdfs[TOKEN_CDF_Q_CTXS][PLANE_TYPES][2][CDF_SIZE(
        7)] = {
      { { { AOM_CDF7(373, 465, 678, 1108, 2134, 5056) },
          { AOM_CDF7(173, 275, 474, 870, 2784, 9771) } },
        { { AOM_CDF7(6280, 8262, 11236, 14340, 18747, 26775) },
          { AOM_CDF7(2057, 3515, 5746, 8425, 13442, 29671) } } },
      { { { AOM_CDF7(1149, 1395, 2137, 3544, 5779, 10343) },
          { AOM_CDF7(294, 474, 838, 1682, 4499, 13540) } },
        { { AOM_CDF7(12776, 15052, 18630, 22072, 25650, 29879) },
          { AOM_CDF7(6210, 10569, 17310, 23755, 27570, 30928) } } },
      { { { AOM_CDF7(3317, 4392, 6615, 10239, 14971, 21621) },
          { AOM_CDF7(588, 1079, 1973, 4055, 8456, 19237) } },
        { { AOM_CDF7(18292, 20058, 23703, 26608, 28853, 31769) },
          { AOM_CDF7(9388, 16293, 23561, 28952, 31072, 32586) } } },
      { { { AOM_CDF7(6759, 8501, 11620, 16030, 21813, 28436) },
          { AOM_CDF7(1208, 2280, 4139, 7332, 13311, 24022) } },
        { { AOM_CDF7(25358, 27110, 30078, 31694, 32475, 32724) },
          { AOM_CDF7(13458, 18725, 24576, 29842, 31598, 32183) } } },
    };

static const aom_cdf_prob
    av1_default_eob_multi128_cdfs[TOKEN_CDF_Q_CTXS][PLANE_TYPES][2][CDF_SIZE(
        8)] = {
      { { { AOM_CDF8(693, 939, 1310, 2060, 3151, 5602, 12620) },
          { AOM_CDF8(114, 206, 391, 1290, 3317, 6071, 12537) } },
        { { AOM_CDF8(9705, 11363, 17511, 19445, 21573, 23355, 28542) },
          { AOM_CDF8(4902, 7998, 12514, 18448, 20512, 21802, 28253) } } },
      { { { AOM_CDF8(1329, 1818, 2514, 3824, 6179, 9616, 17895) },
          { AOM_CDF8(216, 356, 709, 2307, 6058, 10199, 19064) } },
        { { AOM_CDF8(15653, 17962, 22305, 25114, 27331, 29004, 30960) },
          { AOM_CDF8(6679, 11479, 18889, 27341, 29950, 30785, 31516) } } },
      { { { AOM_CDF8(2600, 3508, 5304, 8477, 12607, 17377, 24531) },
          { AOM_CDF8(300, 668, 1470, 4938, 12570, 18869, 25835) } },
        { { AOM_CDF8(18739, 21483, 24858, 27514, 29455, 30605, 31968) },
          { AOM_CDF8(5461, 11099, 18322, 27483, 31535, 32416, 32592) } } },
      { { { AOM_CDF8(4917, 6467, 9434, 14108, 18783, 23990, 29846) },
          { AOM_CDF8(682, 1653, 3738, 11800, 19886, 25614, 30629) } },
        { { AOM_CDF8(24445, 27240, 30036, 31548, 32264, 32555, 32716) },
          { AOM_CDF8(7022, 11703, 21065, 28087, 29257, 30427, 31598) } } },
    };

static const aom_cdf_prob
    av1_default_eob_multi256_cdfs[TOKEN_CDF_Q_CTXS][PLANE_TYPES][2][CDF_SIZE(
        9)] = {
      { { { AOM_CDF9(1911, 2651, 3878, 5131, 6362, 8149, 10305, 15539) },
          { AOM_CDF9(482, 948, 1582, 2688, 6755, 9357, 13182, 18521) } },
        { { AOM_CDF9(6563, 9048, 13129, 14730, 15789, 16556, 17584, 26263) },
          { AOM_CDF9(3040, 4392, 7094, 8445, 12499, 15539, 16215, 22971) } } },
      { { { AOM_CDF9(2341, 2977, 4715, 6995, 10042, 13542, 17070, 22695) },
          { AOM_CDF9(524, 1279, 2241, 3709, 8952, 14201, 20125, 25634) } },
        { { AOM_CDF9(13249, 15850, 19810, 22294, 24265, 25502, 26775, 29692) },
          { AOM_CDF9(6353, 9697, 17053, 22403, 26749, 27084, 27418,
                     30093) } } },
      { { { AOM_CDF9(3832, 5041, 7899, 11629, 16180, 20378, 24470, 28392) },
          { AOM_CDF9(950, 2850, 4165, 5815, 11094, 18188, 25953, 29718) } },
        { { AOM_CDF9(17762, 19980, 23289, 25947, 27924, 28882, 29747, 31546) },
          { AOM_CDF9(7509, 12971, 21163, 25941, 30037, 30720, 31403,
                     32085) } } },
      { { { AOM_CDF9(7219, 9681, 13792, 18500, 22814, 26581, 29566, 31887) },
          { AOM_CDF9(1895, 4116, 6730, 9344, 15584, 23196, 28717, 31690) } },
        { { AOM_CDF9(24807, 26197, 28542, 30832, 31949, 32296, 32476, 32625) },
          { AOM_CDF9(4468, 14895, 23831, 25321, 26810, 28300, 29789,
                     31279) } } },
    };

static const aom_cdf_prob
    av1_default_eob_multi512_cdfs[TOKEN_CDF_Q_CTXS][PLANE_TYPES][2][CDF_SIZE(
        10)] = {
      { { { AOM_CDF10(1022, 2121, 5275, 8179, 10996, 13027, 14783, 17418,
                      22377) },
          { AOM_CDF10(3277, 6554, 9830, 13107, 16384, 19661, 22938, 26214,
                      29491) } },
        { { AOM_CDF10(5208, 6744, 9847, 11629, 12582, 13150, 13580, 13949,
                      30694) },
          { AOM_CDF10(3277, 6554, 9830, 13107, 16384, 19661, 22938, 26214,
                      29491) } } },
      { { { AOM_CDF10(3211, 4958, 7991, 10408, 13976, 16848, 19587, 22712,
                      27627) },
          { AOM_CDF10(3277, 6554, 9830, 13107, 16384, 19661, 22938, 26214,
                      29491) } },
        { { AOM_CDF10(12009, 15222, 20460, 23718, 25624, 26532, 26913, 27440,
                      32577) },
          { AOM_CDF10(3277, 6554, 9830, 13107, 16384, 19661, 22938, 26214,
                      29491) } } },
      { { { AOM_CDF10(4840, 6256, 8940, 12031, 16592, 19797, 23212, 26986,
                      30599) },
          { AOM_CDF10(3277, 6554, 9830, 13107, 16384, 19661, 22938, 26214,
                      29491) } },
        { { AOM_CDF10(15907, 18994, 24689, 27469, 29319, 30136, 30351, 30760,
                      32655) },
          { AOM_CDF10(3277, 6554, 9830, 13107, 16384, 19661, 22938, 26214,
                      29491) } } },
      { { { AOM_CDF10(7344, 9242, 12843, 16504, 21289, 24542, 27923, 30785,
                      32028) },
          { AOM_CDF10(3277, 6554, 9830, 13107, 16384, 19661, 22938, 26214,
                      29491) } },
        { { AOM_CDF10(24120, 26247, 28940, 30682, 32039, 32322, 32464, 32525,
                      32700) },
          { AOM_CDF10(3277, 6554, 9830, 13107, 16384, 19661, 22938, 26214,
                      29491) } } },
    };

static const aom_cdf_prob
    av1_default_eob_multi1024_cdfs[TOKEN_CDF_Q_CTXS][PLANE_TYPES][2][CDF_SIZE(
        11)] = {
      { { { AOM_CDF11(349, 769, 3512, 8048, 13380, 15861, 17645, 18394, 19942,
                      21726) },
          { AOM_CDF11(2979, 5958, 8937, 11916, 14895, 17873, 20852, 23831,
                      26810, 29789) } },
        { { AOM_CDF11(2394, 2967, 5141, 7838, 9658, 10737, 11041, 11142, 11277,
                      29734) },
          { AOM_CDF11(2979, 5958, 8937, 11916, 14895, 17873, 20852, 23831,
                      26810, 29789) } } },
      { { { AOM_CDF11(2546, 3630, 9183, 13423, 18156, 21419, 23412, 24314,
                      25571, 27368) },
          { AOM_CDF11(2979, 5958, 8937, 11916, 14895, 17873, 20852, 23831,
                      26810, 29789) } },
        { { AOM_CDF11(6416, 7552, 12474, 16116, 18183, 19811, 20393, 20536,
                      20661, 32240) },
          { AOM_CDF11(2979, 5958, 8937, 11916, 14895, 17873, 20852, 23831,
                      26810, 29789) } } },
      { { { AOM_CDF11(5102, 6253, 11888, 15776, 20221, 23734, 26178, 27339,
                      28740, 30236) },
          { AOM_CDF11(2979, 5958, 8937, 11916, 14895, 17873, 20852, 23831,
                      26810, 29789) } },
        { { AOM_CDF11(12604, 14163, 20142, 22246, 24028, 25598, 26065, 26199,
                      26243, 32479) },
          { AOM_CDF11(2979, 5958, 8937, 11916, 14895, 17873, 20852, 23831,
                      26810, 29789) } } },
      { { { AOM_CDF11(9285, 10807, 15092, 19816, 23951, 27300, 29275, 30308,
                      31311, 32020) },
          { AOM_CDF11(2979, 5958, 8937, 11916, 14895, 17873, 20852, 23831,
                      26810, 29789) } },
        { { AOM_CDF11(24507, 25696, 28784, 30112, 30961, 31718, 31903, 31965,
                      31996, 32692) },
          { AOM_CDF11(2979, 5958, 8937, 11916, 14895, 17873, 20852, 23831,
                      26810, 29789) } } },
    };

static const aom_cdf_prob av1_default_coeff_base_multi_cdfs
    [TOKEN_CDF_Q_CTXS][TX_SIZES][PLANE_TYPES][SIG_COEF_CONTEXTS]
    [CDF_SIZE(NUM_BASE_LEVELS + 2)] =
        {
          { { { { AOM_CDF4(4462, 11148, 15537) },
                { AOM_CDF4(18264, 30755, 32341) },
                { AOM_CDF4(14122, 27788, 31245) },
                { AOM_CDF4(9572, 22213, 28095) },
                { AOM_CDF4(6129, 16137, 22879) },
                { AOM_CDF4(2396, 6864, 10704) },
                { AOM_CDF4(19658, 30736, 32079) },
                { AOM_CDF4(14361, 27434, 30896) },
                { AOM_CDF4(8665, 20636, 26666) },
                { AOM_CDF4(5279, 14035, 20396) },
                { AOM_CDF4(2691, 7559, 11944) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(16555, 28785, 31272) },
                { AOM_CDF4(12438, 25266, 29768) },
                { AOM_CDF4(5704, 14954, 21411) },
                { AOM_CDF4(3851, 9759, 15119) },
                { AOM_CDF4(2651, 6786, 10985) },
                { AOM_CDF4(19197, 30991, 32196) },
                { AOM_CDF4(15094, 28306, 31286) },
                { AOM_CDF4(8944, 21144, 26942) },
                { AOM_CDF4(4582, 12237, 18765) },
                { AOM_CDF4(1710, 5074, 8401) },
                { AOM_CDF4(22491, 31741, 32377) },
                { AOM_CDF4(15856, 28468, 31274) },
                { AOM_CDF4(8171, 19494, 25316) },
                { AOM_CDF4(3625, 9950, 15314) },
                { AOM_CDF4(1842, 5482, 9080) },
                { AOM_CDF4(23278, 31180, 32046) },
                { AOM_CDF4(15698, 27666, 30720) },
                { AOM_CDF4(6556, 15447, 21080) },
                { AOM_CDF4(2286, 6697, 11415) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(8739, 19251, 23911) },
                { AOM_CDF4(22607, 31690, 32533) },
                { AOM_CDF4(16370, 29189, 31705) },
                { AOM_CDF4(9688, 22103, 27905) },
                { AOM_CDF4(5443, 14551, 21369) },
                { AOM_CDF4(3070, 8678, 13719) },
                { AOM_CDF4(24156, 31889, 32566) },
                { AOM_CDF4(16734, 29422, 31842) },
                { AOM_CDF4(9666, 22190, 28116) },
                { AOM_CDF4(5574, 15559, 22623) },
                { AOM_CDF4(3657, 10421, 16371) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(22073, 31374, 32456) },
                { AOM_CDF4(16997, 29526, 31873) },
                { AOM_CDF4(8534, 20607, 27087) },
                { AOM_CDF4(4771, 14468, 21426) },
                { AOM_CDF4(4089, 11614, 18725) },
                { AOM_CDF4(25249, 32051, 32625) },
                { AOM_CDF4(18701, 30256, 32113) },
                { AOM_CDF4(10557, 23235, 28557) },
                { AOM_CDF4(6309, 15377, 22265) },
                { AOM_CDF4(3217, 9564, 15395) },
                { AOM_CDF4(24835, 31937, 32612) },
                { AOM_CDF4(18659, 30280, 32129) },
                { AOM_CDF4(10696, 22992, 28323) },
                { AOM_CDF4(6242, 15495, 22322) },
                { AOM_CDF4(4446, 9569, 15369) },
                { AOM_CDF4(24935, 31936, 32608) },
                { AOM_CDF4(19146, 30559, 32249) },
                { AOM_CDF4(8802, 20397, 26710) },
                { AOM_CDF4(5435, 10094, 17393) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(9856, 18284, 21827) },
                { AOM_CDF4(26059, 32237, 32565) },
                { AOM_CDF4(19806, 30095, 31706) },
                { AOM_CDF4(11294, 23774, 28530) },
                { AOM_CDF4(6440, 16494, 22901) },
                { AOM_CDF4(2819, 7893, 12211) },
                { AOM_CDF4(26592, 32578, 32720) },
                { AOM_CDF4(22042, 31691, 32466) },
                { AOM_CDF4(14456, 27549, 30942) },
                { AOM_CDF4(8199, 19878, 26163) },
                { AOM_CDF4(3415, 9266, 14132) },
                { AOM_CDF4(23936, 31846, 32470) },
                { AOM_CDF4(16040, 28366, 31207) },
                { AOM_CDF4(9041, 20996, 26680) },
                { AOM_CDF4(5275, 13918, 20301) },
                { AOM_CDF4(2350, 6567, 10463) },
                { AOM_CDF4(22539, 31631, 32400) },
                { AOM_CDF4(16042, 28534, 31190) },
                { AOM_CDF4(9241, 21034, 26794) },
                { AOM_CDF4(5301, 14055, 20372) },
                { AOM_CDF4(2391, 6563, 10442) },
                { AOM_CDF4(25940, 32342, 32690) },
                { AOM_CDF4(22414, 31695, 32502) },
                { AOM_CDF4(15880, 28009, 31099) },
                { AOM_CDF4(9070, 20468, 26695) },
                { AOM_CDF4(4237, 11225, 16915) },
                { AOM_CDF4(25588, 32380, 32632) },
                { AOM_CDF4(19426, 30827, 32189) },
                { AOM_CDF4(11665, 25052, 29617) },
                { AOM_CDF4(6475, 16768, 23598) },
                { AOM_CDF4(2744, 7688, 12023) },
                { AOM_CDF4(27451, 32604, 32723) },
                { AOM_CDF4(21118, 31472, 32401) },
                { AOM_CDF4(12896, 26366, 30360) },
                { AOM_CDF4(7372, 18709, 24965) },
                { AOM_CDF4(3313, 9311, 14250) },
                { AOM_CDF4(27040, 32497, 32670) },
                { AOM_CDF4(20325, 31075, 32278) },
                { AOM_CDF4(11948, 25084, 29512) },
                { AOM_CDF4(7027, 17855, 24267) },
                { AOM_CDF4(3489, 9731, 14876) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(11745, 23067, 26778) },
                { AOM_CDF4(25071, 32147, 32596) },
                { AOM_CDF4(19926, 30514, 31941) },
                { AOM_CDF4(13206, 25647, 29530) },
                { AOM_CDF4(7380, 17863, 24275) },
                { AOM_CDF4(3289, 9424, 13962) },
                { AOM_CDF4(26415, 32555, 32732) },
                { AOM_CDF4(20972, 31488, 32486) },
                { AOM_CDF4(13664, 26742, 30609) },
                { AOM_CDF4(8351, 19980, 26188) },
                { AOM_CDF4(4411, 11626, 17082) },
                { AOM_CDF4(28142, 32455, 32681) },
                { AOM_CDF4(19947, 30749, 32187) },
                { AOM_CDF4(11516, 24090, 29048) },
                { AOM_CDF4(5912, 15814, 22505) },
                { AOM_CDF4(3284, 8872, 13758) },
                { AOM_CDF4(24754, 32134, 32609) },
                { AOM_CDF4(18838, 30537, 32187) },
                { AOM_CDF4(12430, 25578, 30040) },
                { AOM_CDF4(8208, 19343, 25605) },
                { AOM_CDF4(4177, 10963, 16351) },
                { AOM_CDF4(28000, 32534, 32721) },
                { AOM_CDF4(21768, 31544, 32500) },
                { AOM_CDF4(13635, 26637, 30601) },
                { AOM_CDF4(7563, 19062, 25978) },
                { AOM_CDF4(4390, 11935, 18307) },
                { AOM_CDF4(25000, 32431, 32703) },
                { AOM_CDF4(20932, 31526, 32518) },
                { AOM_CDF4(14314, 27946, 31325) },
                { AOM_CDF4(10851, 23360, 28921) },
                { AOM_CDF4(5897, 15520, 22133) },
                { AOM_CDF4(23474, 32064, 32692) },
                { AOM_CDF4(20234, 31075, 32480) },
                { AOM_CDF4(14033, 26970, 31249) },
                { AOM_CDF4(12194, 24453, 29462) },
                { AOM_CDF4(7420, 16583, 21962) },
                { AOM_CDF4(22977, 31982, 32658) },
                { AOM_CDF4(19435, 30913, 32456) },
                { AOM_CDF4(11713, 25342, 30752) },
                { AOM_CDF4(8749, 20236, 27921) },
                { AOM_CDF4(4383, 13804, 20439) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(10755, 20127, 23881) },
                { AOM_CDF4(24354, 31770, 32585) },
                { AOM_CDF4(17309, 29579, 31791) },
                { AOM_CDF4(11556, 24090, 28614) },
                { AOM_CDF4(6193, 16236, 22637) },
                { AOM_CDF4(2455, 6871, 10508) },
                { AOM_CDF4(28585, 32627, 32733) },
                { AOM_CDF4(21783, 31704, 32512) },
                { AOM_CDF4(14206, 27735, 31142) },
                { AOM_CDF4(8579, 20749, 26779) },
                { AOM_CDF4(3344, 9122, 13657) },
                { AOM_CDF4(29568, 32604, 32711) },
                { AOM_CDF4(21312, 31184, 32268) },
                { AOM_CDF4(12770, 25999, 30141) },
                { AOM_CDF4(7652, 19125, 25430) },
                { AOM_CDF4(3286, 9392, 14319) },
                { AOM_CDF4(27758, 32539, 32700) },
                { AOM_CDF4(20811, 31421, 32413) },
                { AOM_CDF4(14047, 27753, 31196) },
                { AOM_CDF4(9015, 21723, 27608) },
                { AOM_CDF4(3879, 10482, 15686) },
                { AOM_CDF4(29707, 32711, 32748) },
                { AOM_CDF4(23194, 32086, 32635) },
                { AOM_CDF4(15310, 29109, 31927) },
                { AOM_CDF4(10241, 23995, 29594) },
                { AOM_CDF4(5171, 13799, 19770) },
                { AOM_CDF4(28022, 32588, 32705) },
                { AOM_CDF4(20949, 31502, 32369) },
                { AOM_CDF4(12876, 25966, 30093) },
                { AOM_CDF4(7989, 19482, 25860) },
                { AOM_CDF4(3673, 10233, 15494) },
                { AOM_CDF4(29706, 32686, 32747) },
                { AOM_CDF4(23091, 32054, 32605) },
                { AOM_CDF4(14800, 28751, 31652) },
                { AOM_CDF4(9850, 23230, 28722) },
                { AOM_CDF4(4536, 12001, 17689) },
                { AOM_CDF4(29400, 32698, 32748) },
                { AOM_CDF4(23285, 32109, 32653) },
                { AOM_CDF4(15287, 29288, 32016) },
                { AOM_CDF4(10640, 24648, 29857) },
                { AOM_CDF4(5243, 13730, 19559) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(9521, 21806, 26977) },
                { AOM_CDF4(27936, 32497, 32698) },
                { AOM_CDF4(18881, 30706, 32224) },
                { AOM_CDF4(13866, 27418, 30762) },
                { AOM_CDF4(9859, 22155, 26794) },
                { AOM_CDF4(2721, 7754, 12153) },
                { AOM_CDF4(28128, 32634, 32742) },
                { AOM_CDF4(21600, 31722, 32542) },
                { AOM_CDF4(15041, 28744, 31513) },
                { AOM_CDF4(9793, 23312, 28915) },
                { AOM_CDF4(4200, 10874, 15808) },
                { AOM_CDF4(30338, 32581, 32730) },
                { AOM_CDF4(19260, 30457, 32025) },
                { AOM_CDF4(11154, 23294, 28333) },
                { AOM_CDF4(6379, 16048, 22696) },
                { AOM_CDF4(3584, 8558, 13751) },
                { AOM_CDF4(28828, 32660, 32748) },
                { AOM_CDF4(22611, 32073, 32640) },
                { AOM_CDF4(18081, 30085, 32173) },
                { AOM_CDF4(13957, 26831, 30432) },
                { AOM_CDF4(10019, 20420, 24576) },
                { AOM_CDF4(30576, 32671, 32748) },
                { AOM_CDF4(22696, 31593, 32485) },
                { AOM_CDF4(15674, 28435, 31466) },
                { AOM_CDF4(11781, 24480, 29491) },
                { AOM_CDF4(6809, 15880, 21918) },
                { AOM_CDF4(21357, 31843, 32614) },
                { AOM_CDF4(14847, 30213, 32361) },
                { AOM_CDF4(13543, 28177, 31907) },
                { AOM_CDF4(12702, 26785, 31111) },
                { AOM_CDF4(7193, 16784, 24456) },
                { AOM_CDF4(21005, 31928, 32348) },
                { AOM_CDF4(19183, 30720, 32358) },
                { AOM_CDF4(17324, 29421, 31652) },
                { AOM_CDF4(13107, 27088, 30693) },
                { AOM_CDF4(8001, 19813, 25338) },
                { AOM_CDF4(19583, 31038, 32508) },
                { AOM_CDF4(15796, 29040, 32034) },
                { AOM_CDF4(11833, 25767, 31112) },
                { AOM_CDF4(10352, 25257, 30550) },
                { AOM_CDF4(6716, 17910, 25339) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(7063, 14444, 18604) },
                { AOM_CDF4(15912, 24676, 26941) },
                { AOM_CDF4(9191, 20014, 23665) },
                { AOM_CDF4(8128, 18536, 23098) },
                { AOM_CDF4(4206, 11237, 16761) },
                { AOM_CDF4(1448, 3854, 6750) },
                { AOM_CDF4(28852, 32444, 32697) },
                { AOM_CDF4(19419, 30257, 31943) },
                { AOM_CDF4(10845, 23636, 28172) },
                { AOM_CDF4(5734, 15504, 21549) },
                { AOM_CDF4(2102, 6549, 10148) },
                { AOM_CDF4(30395, 32493, 32664) },
                { AOM_CDF4(20186, 30719, 32099) },
                { AOM_CDF4(12179, 25108, 29492) },
                { AOM_CDF4(7320, 19063, 25080) },
                { AOM_CDF4(2823, 8258, 12779) },
                { AOM_CDF4(28527, 32451, 32659) },
                { AOM_CDF4(20969, 31204, 32337) },
                { AOM_CDF4(14031, 27804, 30992) },
                { AOM_CDF4(8371, 20830, 26494) },
                { AOM_CDF4(3772, 10138, 15466) },
                { AOM_CDF4(31164, 32732, 32748) },
                { AOM_CDF4(24234, 32137, 32678) },
                { AOM_CDF4(15441, 29486, 32102) },
                { AOM_CDF4(10767, 25109, 30246) },
                { AOM_CDF4(4921, 13255, 18719) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(14229, 25859, 29291) },
                { AOM_CDF4(28574, 32133, 32634) },
                { AOM_CDF4(18505, 28765, 31151) },
                { AOM_CDF4(12358, 25091, 29023) },
                { AOM_CDF4(8623, 18971, 26732) },
                { AOM_CDF4(2621, 5243, 9175) },
                { AOM_CDF4(30500, 32742, 32748) },
                { AOM_CDF4(24685, 32282, 32693) },
                { AOM_CDF4(16954, 29686, 31897) },
                { AOM_CDF4(7473, 19546, 22995) },
                { AOM_CDF4(2341, 8192, 12873) },
                { AOM_CDF4(32083, 32752, 32756) },
                { AOM_CDF4(23395, 31504, 32315) },
                { AOM_CDF4(12023, 24046, 28525) },
                { AOM_CDF4(3641, 13350, 21845) },
                { AOM_CDF4(7085, 11513, 15056) },
                { AOM_CDF4(30422, 32727, 32748) },
                { AOM_CDF4(23632, 32381, 32742) },
                { AOM_CDF4(17847, 30949, 32571) },
                { AOM_CDF4(11785, 27019, 31618) },
                { AOM_CDF4(1365, 4096, 10923) },
                { AOM_CDF4(32069, 32731, 32748) },
                { AOM_CDF4(25871, 31959, 32471) },
                { AOM_CDF4(18445, 29287, 31531) },
                { AOM_CDF4(15477, 26083, 29941) },
                { AOM_CDF4(9760, 18135, 23552) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(5206, 12013, 15839) },
                { AOM_CDF4(25287, 29818, 31398) },
                { AOM_CDF4(8979, 17375, 20640) },
                { AOM_CDF4(3053, 8955, 12619) },
                { AOM_CDF4(3390, 9039, 13277) },
                { AOM_CDF4(2216, 6064, 9795) },
                { AOM_CDF4(28440, 32040, 32642) },
                { AOM_CDF4(16150, 28157, 31565) },
                { AOM_CDF4(8521, 19958, 26821) },
                { AOM_CDF4(3760, 14182, 20198) },
                { AOM_CDF4(1990, 8358, 12338) },
                { AOM_CDF4(29744, 32239, 32687) },
                { AOM_CDF4(19739, 30444, 32246) },
                { AOM_CDF4(9786, 23171, 29043) },
                { AOM_CDF4(4151, 15292, 21845) },
                { AOM_CDF4(1872, 6366, 9175) },
                { AOM_CDF4(28893, 32552, 32744) },
                { AOM_CDF4(20004, 31355, 32565) },
                { AOM_CDF4(12806, 27024, 30828) },
                { AOM_CDF4(6351, 17161, 23579) },
                { AOM_CDF4(3296, 8337, 12700) },
                { AOM_CDF4(32272, 32752, 32756) },
                { AOM_CDF4(26918, 32469, 32717) },
                { AOM_CDF4(16886, 29126, 31860) },
                { AOM_CDF4(7927, 19085, 27012) },
                { AOM_CDF4(4274, 11462, 18651) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } } },
          { { { { AOM_CDF4(4124, 11059, 15633) },
                { AOM_CDF4(18394, 30703, 32248) },
                { AOM_CDF4(13398, 27103, 30987) },
                { AOM_CDF4(8665, 20604, 26909) },
                { AOM_CDF4(5309, 14230, 21005) },
                { AOM_CDF4(2359, 6454, 10409) },
                { AOM_CDF4(21026, 31227, 32281) },
                { AOM_CDF4(14136, 27385, 31026) },
                { AOM_CDF4(7712, 19722, 26214) },
                { AOM_CDF4(4688, 13105, 19899) },
                { AOM_CDF4(2726, 7811, 12613) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(17825, 29738, 31751) },
                { AOM_CDF4(12572, 25682, 30276) },
                { AOM_CDF4(5725, 15610, 22799) },
                { AOM_CDF4(4377, 10950, 17689) },
                { AOM_CDF4(3170, 8066, 12758) },
                { AOM_CDF4(21492, 31555, 32375) },
                { AOM_CDF4(15141, 28588, 31403) },
                { AOM_CDF4(8790, 20480, 26487) },
                { AOM_CDF4(3765, 10836, 17481) },
                { AOM_CDF4(1671, 4757, 8206) },
                { AOM_CDF4(24691, 32136, 32502) },
                { AOM_CDF4(16369, 28517, 31232) },
                { AOM_CDF4(8516, 19478, 25157) },
                { AOM_CDF4(3340, 9884, 15638) },
                { AOM_CDF4(2039, 5664, 9589) },
                { AOM_CDF4(26997, 31917, 32381) },
                { AOM_CDF4(17402, 28370, 31028) },
                { AOM_CDF4(7668, 17256, 23130) },
                { AOM_CDF4(2476, 7667, 13098) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(7807, 18182, 23504) },
                { AOM_CDF4(22227, 31588, 32513) },
                { AOM_CDF4(14304, 28166, 31501) },
                { AOM_CDF4(8350, 20565, 27310) },
                { AOM_CDF4(5048, 14009, 21308) },
                { AOM_CDF4(3477, 9406, 15089) },
                { AOM_CDF4(24745, 32070, 32638) },
                { AOM_CDF4(15554, 29413, 32047) },
                { AOM_CDF4(9228, 22656, 28919) },
                { AOM_CDF4(5706, 16412, 24045) },
                { AOM_CDF4(4691, 12165, 18561) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(23283, 31916, 32660) },
                { AOM_CDF4(17021, 30258, 32297) },
                { AOM_CDF4(10345, 23603, 29452) },
                { AOM_CDF4(3968, 17792, 25088) },
                { AOM_CDF4(2260, 12429, 24858) },
                { AOM_CDF4(26285, 32108, 32658) },
                { AOM_CDF4(18293, 29887, 32051) },
                { AOM_CDF4(9282, 21866, 27783) },
                { AOM_CDF4(4479, 12129, 20937) },
                { AOM_CDF4(2453, 8830, 14226) },
                { AOM_CDF4(25781, 32080, 32618) },
                { AOM_CDF4(18779, 29929, 32165) },
                { AOM_CDF4(10923, 22578, 27809) },
                { AOM_CDF4(5712, 12927, 21044) },
                { AOM_CDF4(3955, 12429, 18079) },
                { AOM_CDF4(25820, 31951, 32604) },
                { AOM_CDF4(19699, 30537, 32299) },
                { AOM_CDF4(9862, 19724, 27360) },
                { AOM_CDF4(4681, 9362, 14043) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(6566, 14440, 19033) },
                { AOM_CDF4(22455, 31575, 32407) },
                { AOM_CDF4(15831, 28483, 31329) },
                { AOM_CDF4(9693, 22225, 27827) },
                { AOM_CDF4(5703, 15240, 22125) },
                { AOM_CDF4(2543, 7444, 11683) },
                { AOM_CDF4(25817, 32497, 32694) },
                { AOM_CDF4(19968, 31041, 32317) },
                { AOM_CDF4(12383, 25712, 30150) },
                { AOM_CDF4(6777, 17647, 24575) },
                { AOM_CDF4(3137, 8905, 13866) },
                { AOM_CDF4(24846, 31915, 32452) },
                { AOM_CDF4(15632, 28234, 31232) },
                { AOM_CDF4(8423, 20400, 26656) },
                { AOM_CDF4(5175, 13906, 20512) },
                { AOM_CDF4(2479, 7115, 11521) },
                { AOM_CDF4(23211, 31770, 32458) },
                { AOM_CDF4(15633, 28161, 31186) },
                { AOM_CDF4(7969, 19692, 26070) },
                { AOM_CDF4(4835, 13326, 20018) },
                { AOM_CDF4(2515, 6981, 11176) },
                { AOM_CDF4(26779, 32565, 32704) },
                { AOM_CDF4(22101, 31714, 32494) },
                { AOM_CDF4(14543, 27410, 30903) },
                { AOM_CDF4(7657, 19345, 26054) },
                { AOM_CDF4(3947, 11234, 17307) },
                { AOM_CDF4(25261, 32068, 32576) },
                { AOM_CDF4(17540, 29996, 31979) },
                { AOM_CDF4(10592, 23809, 28945) },
                { AOM_CDF4(5871, 15705, 22341) },
                { AOM_CDF4(2602, 7262, 11582) },
                { AOM_CDF4(28230, 32583, 32709) },
                { AOM_CDF4(20517, 31175, 32331) },
                { AOM_CDF4(12846, 26061, 30040) },
                { AOM_CDF4(7118, 18243, 24556) },
                { AOM_CDF4(3129, 9001, 14174) },
                { AOM_CDF4(28850, 32619, 32713) },
                { AOM_CDF4(20949, 31348, 32384) },
                { AOM_CDF4(12391, 25575, 29806) },
                { AOM_CDF4(6856, 17474, 23996) },
                { AOM_CDF4(3312, 9642, 14989) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(9723, 20276, 24593) },
                { AOM_CDF4(24514, 31812, 32479) },
                { AOM_CDF4(18048, 29628, 31519) },
                { AOM_CDF4(11589, 23830, 28596) },
                { AOM_CDF4(6357, 16468, 23195) },
                { AOM_CDF4(3134, 8999, 13786) },
                { AOM_CDF4(26258, 32407, 32712) },
                { AOM_CDF4(19022, 30642, 32264) },
                { AOM_CDF4(10810, 24078, 29283) },
                { AOM_CDF4(5904, 16192, 23183) },
                { AOM_CDF4(3581, 10547, 16335) },
                { AOM_CDF4(27887, 32385, 32671) },
                { AOM_CDF4(17809, 29882, 32038) },
                { AOM_CDF4(9580, 21976, 27833) },
                { AOM_CDF4(5880, 15087, 21588) },
                { AOM_CDF4(3255, 8941, 14428) },
                { AOM_CDF4(24471, 32058, 32603) },
                { AOM_CDF4(17012, 29990, 32090) },
                { AOM_CDF4(10112, 23199, 29008) },
                { AOM_CDF4(5836, 16020, 23207) },
                { AOM_CDF4(3503, 9946, 15663) },
                { AOM_CDF4(28320, 32550, 32732) },
                { AOM_CDF4(20066, 31287, 32503) },
                { AOM_CDF4(11948, 25973, 30537) },
                { AOM_CDF4(7074, 18828, 26476) },
                { AOM_CDF4(4582, 12635, 19990) },
                { AOM_CDF4(25121, 32151, 32666) },
                { AOM_CDF4(18222, 30616, 32285) },
                { AOM_CDF4(11542, 25440, 30517) },
                { AOM_CDF4(8306, 20537, 26685) },
                { AOM_CDF4(4800, 15445, 22124) },
                { AOM_CDF4(22309, 31899, 32690) },
                { AOM_CDF4(18121, 30818, 32484) },
                { AOM_CDF4(12285, 26556, 30977) },
                { AOM_CDF4(10198, 21734, 27920) },
                { AOM_CDF4(4228, 13213, 20612) },
                { AOM_CDF4(21103, 31836, 32696) },
                { AOM_CDF4(18836, 31086, 32533) },
                { AOM_CDF4(12013, 26071, 30877) },
                { AOM_CDF4(8326, 17996, 26322) },
                { AOM_CDF4(4599, 12072, 18396) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(9550, 17624, 21538) },
                { AOM_CDF4(24882, 31701, 32513) },
                { AOM_CDF4(15356, 28381, 31130) },
                { AOM_CDF4(9747, 21795, 27083) },
                { AOM_CDF4(5300, 14553, 21009) },
                { AOM_CDF4(2247, 6514, 10325) },
                { AOM_CDF4(28009, 32592, 32719) },
                { AOM_CDF4(20378, 31014, 32360) },
                { AOM_CDF4(11639, 24941, 29911) },
                { AOM_CDF4(6649, 17900, 24697) },
                { AOM_CDF4(2812, 7993, 12458) },
                { AOM_CDF4(29842, 32574, 32710) },
                { AOM_CDF4(20150, 30644, 32145) },
                { AOM_CDF4(11176, 23964, 28983) },
                { AOM_CDF4(6463, 16904, 23591) },
                { AOM_CDF4(2983, 8898, 13695) },
                { AOM_CDF4(27690, 32557, 32697) },
                { AOM_CDF4(21125, 31239, 32357) },
                { AOM_CDF4(12439, 25456, 30046) },
                { AOM_CDF4(6951, 18075, 24861) },
                { AOM_CDF4(3043, 8862, 13681) },
                { AOM_CDF4(30454, 32729, 32748) },
                { AOM_CDF4(24373, 32272, 32661) },
                { AOM_CDF4(15239, 28610, 31667) },
                { AOM_CDF4(8752, 21867, 28297) },
                { AOM_CDF4(4481, 12487, 18574) },
                { AOM_CDF4(27738, 32478, 32675) },
                { AOM_CDF4(19407, 30740, 32198) },
                { AOM_CDF4(11116, 23705, 28702) },
                { AOM_CDF4(5753, 15450, 21848) },
                { AOM_CDF4(2846, 8210, 12891) },
                { AOM_CDF4(29962, 32670, 32736) },
                { AOM_CDF4(22991, 31878, 32529) },
                { AOM_CDF4(14477, 27576, 30994) },
                { AOM_CDF4(7999, 19669, 26297) },
                { AOM_CDF4(3670, 10611, 16337) },
                { AOM_CDF4(30784, 32735, 32748) },
                { AOM_CDF4(24678, 32435, 32701) },
                { AOM_CDF4(16718, 29725, 32014) },
                { AOM_CDF4(9547, 22615, 28356) },
                { AOM_CDF4(4355, 12438, 18375) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(7427, 17188, 21890) },
                { AOM_CDF4(26814, 32222, 32608) },
                { AOM_CDF4(16461, 28592, 31381) },
                { AOM_CDF4(9778, 22398, 27133) },
                { AOM_CDF4(5815, 14867, 20328) },
                { AOM_CDF4(2178, 6347, 9744) },
                { AOM_CDF4(28035, 32440, 32725) },
                { AOM_CDF4(18970, 30651, 32293) },
                { AOM_CDF4(11805, 25835, 29912) },
                { AOM_CDF4(7871, 19918, 26147) },
                { AOM_CDF4(3369, 9073, 13929) },
                { AOM_CDF4(29403, 32382, 32627) },
                { AOM_CDF4(15885, 28037, 31077) },
                { AOM_CDF4(8006, 19228, 25221) },
                { AOM_CDF4(4977, 13204, 19080) },
                { AOM_CDF4(2887, 8504, 12795) },
                { AOM_CDF4(28805, 32541, 32734) },
                { AOM_CDF4(19886, 30860, 32365) },
                { AOM_CDF4(13722, 26501, 30560) },
                { AOM_CDF4(8960, 21018, 26432) },
                { AOM_CDF4(3707, 10612, 16331) },
                { AOM_CDF4(30766, 32712, 32748) },
                { AOM_CDF4(21700, 31760, 32587) },
                { AOM_CDF4(14844, 28410, 31602) },
                { AOM_CDF4(8942, 21893, 27890) },
                { AOM_CDF4(4622, 12781, 19033) },
                { AOM_CDF4(24181, 32218, 32631) },
                { AOM_CDF4(11995, 30876, 32087) },
                { AOM_CDF4(13073, 27335, 30900) },
                { AOM_CDF4(12156, 25897, 31711) },
                { AOM_CDF4(8192, 15019, 20480) },
                { AOM_CDF4(17695, 31457, 32113) },
                { AOM_CDF4(15750, 30414, 32406) },
                { AOM_CDF4(15278, 28546, 31160) },
                { AOM_CDF4(10169, 27683, 30508) },
                { AOM_CDF4(19661, 26214, 29491) },
                { AOM_CDF4(20361, 31225, 32649) },
                { AOM_CDF4(18018, 31367, 32631) },
                { AOM_CDF4(13879, 28503, 32362) },
                { AOM_CDF4(11108, 28325, 31380) },
                { AOM_CDF4(10923, 26526, 31208) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(5556, 12459, 16671) },
                { AOM_CDF4(19510, 28933, 30949) },
                { AOM_CDF4(10782, 22318, 26991) },
                { AOM_CDF4(6425, 17337, 23417) },
                { AOM_CDF4(3353, 9760, 15789) },
                { AOM_CDF4(1378, 3974, 6683) },
                { AOM_CDF4(27885, 32206, 32661) },
                { AOM_CDF4(16506, 28800, 31566) },
                { AOM_CDF4(8880, 21100, 26867) },
                { AOM_CDF4(4925, 14039, 20589) },
                { AOM_CDF4(2003, 5812, 9458) },
                { AOM_CDF4(29532, 32323, 32578) },
                { AOM_CDF4(18719, 30254, 31943) },
                { AOM_CDF4(10881, 23591, 28851) },
                { AOM_CDF4(6147, 16595, 23036) },
                { AOM_CDF4(2871, 7918, 12170) },
                { AOM_CDF4(29145, 32568, 32707) },
                { AOM_CDF4(20594, 31091, 32274) },
                { AOM_CDF4(12739, 25878, 29947) },
                { AOM_CDF4(7180, 17988, 24861) },
                { AOM_CDF4(3177, 8838, 13418) },
                { AOM_CDF4(31541, 32728, 32748) },
                { AOM_CDF4(22662, 32095, 32666) },
                { AOM_CDF4(15674, 29774, 32154) },
                { AOM_CDF4(10785, 24760, 29924) },
                { AOM_CDF4(4240, 11619, 16898) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(11702, 20790, 25601) },
                { AOM_CDF4(28677, 31964, 32629) },
                { AOM_CDF4(11705, 25802, 29972) },
                { AOM_CDF4(6554, 20637, 26214) },
                { AOM_CDF4(4896, 13559, 19209) },
                { AOM_CDF4(923, 5307, 8769) },
                { AOM_CDF4(31551, 32717, 32748) },
                { AOM_CDF4(19904, 30257, 32020) },
                { AOM_CDF4(9567, 21287, 27028) },
                { AOM_CDF4(3948, 15002, 19345) },
                { AOM_CDF4(2328, 6983, 9848) },
                { AOM_CDF4(31958, 32705, 32748) },
                { AOM_CDF4(15737, 28682, 31397) },
                { AOM_CDF4(7115, 19661, 25653) },
                { AOM_CDF4(3855, 16384, 24094) },
                { AOM_CDF4(6899, 10348, 16384) },
                { AOM_CDF4(31573, 32718, 32748) },
                { AOM_CDF4(22945, 31225, 32432) },
                { AOM_CDF4(13720, 25975, 31170) },
                { AOM_CDF4(10042, 18498, 24312) },
                { AOM_CDF4(4024, 7473, 12072) },
                { AOM_CDF4(32537, 32744, 32748) },
                { AOM_CDF4(26153, 31795, 32497) },
                { AOM_CDF4(19474, 28984, 31601) },
                { AOM_CDF4(15023, 25490, 30449) },
                { AOM_CDF4(7205, 14844, 22131) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(3024, 7914, 11762) },
                { AOM_CDF4(18244, 25626, 28987) },
                { AOM_CDF4(7226, 16180, 20955) },
                { AOM_CDF4(3410, 10738, 16315) },
                { AOM_CDF4(4182, 10948, 16650) },
                { AOM_CDF4(1810, 5929, 10591) },
                { AOM_CDF4(26369, 31769, 32629) },
                { AOM_CDF4(14376, 27825, 31517) },
                { AOM_CDF4(8612, 21055, 27222) },
                { AOM_CDF4(5249, 14981, 21614) },
                { AOM_CDF4(2690, 9113, 13791) },
                { AOM_CDF4(28616, 32032, 32563) },
                { AOM_CDF4(18555, 29802, 31826) },
                { AOM_CDF4(9396, 22063, 27919) },
                { AOM_CDF4(4859, 13290, 20018) },
                { AOM_CDF4(2231, 7530, 10737) },
                { AOM_CDF4(29866, 32582, 32722) },
                { AOM_CDF4(20920, 30842, 32247) },
                { AOM_CDF4(11731, 25256, 29932) },
                { AOM_CDF4(6601, 16994, 23940) },
                { AOM_CDF4(3530, 9640, 14946) },
                { AOM_CDF4(32170, 32751, 32755) },
                { AOM_CDF4(26498, 32236, 32685) },
                { AOM_CDF4(13492, 27236, 31180) },
                { AOM_CDF4(7516, 19243, 26802) },
                { AOM_CDF4(4444, 11879, 18300) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } } },
          { { { { AOM_CDF4(4002, 11946, 17010) },
                { AOM_CDF4(19950, 30806, 32374) },
                { AOM_CDF4(12490, 26525, 30902) },
                { AOM_CDF4(7545, 19161, 26353) },
                { AOM_CDF4(4751, 12894, 20380) },
                { AOM_CDF4(2483, 7446, 12167) },
                { AOM_CDF4(22074, 31614, 32455) },
                { AOM_CDF4(13304, 27337, 31389) },
                { AOM_CDF4(7190, 19638, 26814) },
                { AOM_CDF4(4623, 13399, 20891) },
                { AOM_CDF4(3414, 9185, 14944) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(19898, 30891, 32341) },
                { AOM_CDF4(13312, 27172, 31225) },
                { AOM_CDF4(6872, 17894, 25643) },
                { AOM_CDF4(6330, 12884, 20655) },
                { AOM_CDF4(5925, 11003, 14873) },
                { AOM_CDF4(23872, 31922, 32413) },
                { AOM_CDF4(14591, 28022, 31251) },
                { AOM_CDF4(7800, 18666, 25337) },
                { AOM_CDF4(2933, 9493, 15953) },
                { AOM_CDF4(1377, 4555, 8608) },
                { AOM_CDF4(27231, 32205, 32509) },
                { AOM_CDF4(16691, 27977, 31050) },
                { AOM_CDF4(8922, 19309, 25077) },
                { AOM_CDF4(3629, 11017, 16520) },
                { AOM_CDF4(2392, 6578, 10763) },
                { AOM_CDF4(30006, 32359, 32570) },
                { AOM_CDF4(18797, 28390, 31221) },
                { AOM_CDF4(8526, 19522, 25646) },
                { AOM_CDF4(2365, 8783, 15877) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(7361, 18287, 24405) },
                { AOM_CDF4(22334, 31620, 32570) },
                { AOM_CDF4(13357, 28299, 31763) },
                { AOM_CDF4(8540, 21286, 27979) },
                { AOM_CDF4(5260, 15486, 23131) },
                { AOM_CDF4(3197, 8743, 15040) },
                { AOM_CDF4(25589, 32327, 32707) },
                { AOM_CDF4(15905, 30232, 32380) },
                { AOM_CDF4(9900, 24007, 30031) },
                { AOM_CDF4(6115, 16462, 24641) },
                { AOM_CDF4(4145, 10879, 18392) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(25799, 32564, 32748) },
                { AOM_CDF4(18861, 31353, 32622) },
                { AOM_CDF4(11786, 26422, 31473) },
                { AOM_CDF4(8192, 21845, 27307) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(27314, 32139, 32687) },
                { AOM_CDF4(17455, 29104, 31894) },
                { AOM_CDF4(8392, 19021, 26134) },
                { AOM_CDF4(7142, 13863, 19745) },
                { AOM_CDF4(4274, 5699, 9973) },
                { AOM_CDF4(26484, 31988, 32579) },
                { AOM_CDF4(18920, 29829, 32084) },
                { AOM_CDF4(11118, 22821, 29842) },
                { AOM_CDF4(5783, 11565, 19275) },
                { AOM_CDF4(7282, 18204, 25486) },
                { AOM_CDF4(26862, 31879, 32514) },
                { AOM_CDF4(19780, 30438, 32421) },
                { AOM_CDF4(2731, 15019, 27307) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(5306, 13172, 18023) },
                { AOM_CDF4(22174, 31768, 32544) },
                { AOM_CDF4(14504, 28207, 31587) },
                { AOM_CDF4(9063, 21603, 27790) },
                { AOM_CDF4(5747, 15142, 22018) },
                { AOM_CDF4(2665, 7998, 12578) },
                { AOM_CDF4(25826, 32442, 32695) },
                { AOM_CDF4(17050, 29997, 32168) },
                { AOM_CDF4(9931, 23538, 29315) },
                { AOM_CDF4(6323, 16989, 24177) },
                { AOM_CDF4(3544, 10012, 15786) },
                { AOM_CDF4(24636, 31907, 32551) },
                { AOM_CDF4(14485, 27834, 31354) },
                { AOM_CDF4(8382, 20539, 26960) },
                { AOM_CDF4(5196, 14452, 21368) },
                { AOM_CDF4(2976, 8504, 13546) },
                { AOM_CDF4(23187, 31744, 32515) },
                { AOM_CDF4(14608, 27811, 31297) },
                { AOM_CDF4(7810, 20056, 26742) },
                { AOM_CDF4(5206, 14392, 21258) },
                { AOM_CDF4(3037, 8791, 13853) },
                { AOM_CDF4(27745, 32568, 32716) },
                { AOM_CDF4(19537, 31025, 32427) },
                { AOM_CDF4(11275, 25278, 30482) },
                { AOM_CDF4(7104, 18978, 26240) },
                { AOM_CDF4(4434, 12675, 19421) },
                { AOM_CDF4(26891, 32123, 32590) },
                { AOM_CDF4(16499, 29303, 31734) },
                { AOM_CDF4(9128, 21581, 27348) },
                { AOM_CDF4(4831, 13689, 20215) },
                { AOM_CDF4(2299, 6583, 10905) },
                { AOM_CDF4(29253, 32565, 32702) },
                { AOM_CDF4(19375, 30467, 32129) },
                { AOM_CDF4(11245, 24302, 29078) },
                { AOM_CDF4(6046, 16407, 23539) },
                { AOM_CDF4(3088, 9440, 15273) },
                { AOM_CDF4(30400, 32645, 32733) },
                { AOM_CDF4(20688, 30891, 32312) },
                { AOM_CDF4(11007, 24065, 29305) },
                { AOM_CDF4(6289, 17016, 24026) },
                { AOM_CDF4(3628, 10533, 16695) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(8646, 19108, 23905) },
                { AOM_CDF4(23739, 31943, 32479) },
                { AOM_CDF4(16012, 29063, 31413) },
                { AOM_CDF4(11382, 23805, 28343) },
                { AOM_CDF4(5585, 15795, 22502) },
                { AOM_CDF4(4021, 10454, 16552) },
                { AOM_CDF4(25774, 32257, 32715) },
                { AOM_CDF4(17392, 30186, 32209) },
                { AOM_CDF4(9999, 23407, 29063) },
                { AOM_CDF4(5889, 16415, 23982) },
                { AOM_CDF4(3924, 11892, 18535) },
                { AOM_CDF4(28039, 32395, 32654) },
                { AOM_CDF4(15958, 29493, 31882) },
                { AOM_CDF4(10150, 22545, 28351) },
                { AOM_CDF4(6983, 16262, 22806) },
                { AOM_CDF4(4736, 10923, 17689) },
                { AOM_CDF4(25247, 32153, 32644) },
                { AOM_CDF4(16540, 30064, 32169) },
                { AOM_CDF4(9724, 23041, 28984) },
                { AOM_CDF4(6087, 16012, 23510) },
                { AOM_CDF4(3636, 11238, 18465) },
                { AOM_CDF4(28641, 32636, 32745) },
                { AOM_CDF4(19619, 31455, 32600) },
                { AOM_CDF4(11576, 26365, 31068) },
                { AOM_CDF4(6688, 19832, 28387) },
                { AOM_CDF4(6425, 14537, 22327) },
                { AOM_CDF4(24312, 32051, 32684) },
                { AOM_CDF4(17971, 29948, 32148) },
                { AOM_CDF4(11327, 23868, 29329) },
                { AOM_CDF4(5174, 13797, 20696) },
                { AOM_CDF4(3277, 13107, 22938) },
                { AOM_CDF4(23396, 31785, 32571) },
                { AOM_CDF4(19640, 31350, 32348) },
                { AOM_CDF4(14043, 19505, 26526) },
                { AOM_CDF4(4681, 14043, 18725) },
                { AOM_CDF4(14043, 23406, 28087) },
                { AOM_CDF4(22952, 32046, 32624) },
                { AOM_CDF4(18053, 30666, 32521) },
                { AOM_CDF4(14336, 24576, 28672) },
                { AOM_CDF4(4681, 14043, 18725) },
                { AOM_CDF4(6554, 13107, 19661) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(6447, 14363, 18951) },
                { AOM_CDF4(23736, 31444, 32532) },
                { AOM_CDF4(13480, 27014, 30790) },
                { AOM_CDF4(8395, 20532, 26629) },
                { AOM_CDF4(5255, 14331, 21040) },
                { AOM_CDF4(2289, 7037, 11241) },
                { AOM_CDF4(27036, 32527, 32712) },
                { AOM_CDF4(17223, 30108, 32192) },
                { AOM_CDF4(10225, 23887, 29577) },
                { AOM_CDF4(6309, 17098, 24150) },
                { AOM_CDF4(2919, 8483, 13233) },
                { AOM_CDF4(29436, 32530, 32712) },
                { AOM_CDF4(18119, 29975, 32059) },
                { AOM_CDF4(10039, 22952, 28783) },
                { AOM_CDF4(6075, 16199, 23020) },
                { AOM_CDF4(3176, 9307, 14620) },
                { AOM_CDF4(26945, 32388, 32672) },
                { AOM_CDF4(17686, 29999, 32134) },
                { AOM_CDF4(10104, 23162, 29005) },
                { AOM_CDF4(6120, 16494, 23597) },
                { AOM_CDF4(2941, 8550, 13739) },
                { AOM_CDF4(30723, 32723, 32748) },
                { AOM_CDF4(23261, 32041, 32636) },
                { AOM_CDF4(13607, 27654, 31409) },
                { AOM_CDF4(7944, 20700, 27589) },
                { AOM_CDF4(4415, 12702, 19331) },
                { AOM_CDF4(27993, 32319, 32650) },
                { AOM_CDF4(16599, 29380, 31793) },
                { AOM_CDF4(8660, 20044, 26077) },
                { AOM_CDF4(4517, 12317, 18513) },
                { AOM_CDF4(2357, 6808, 11259) },
                { AOM_CDF4(30290, 32609, 32712) },
                { AOM_CDF4(21059, 31180, 32308) },
                { AOM_CDF4(11912, 24987, 29897) },
                { AOM_CDF4(6737, 17747, 25009) },
                { AOM_CDF4(3386, 10886, 16566) },
                { AOM_CDF4(31525, 32732, 32748) },
                { AOM_CDF4(24173, 32214, 32664) },
                { AOM_CDF4(14606, 27824, 31289) },
                { AOM_CDF4(7942, 19819, 26273) },
                { AOM_CDF4(4666, 12492, 18643) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(6767, 15371, 20330) },
                { AOM_CDF4(25254, 31761, 32478) },
                { AOM_CDF4(14651, 27133, 30714) },
                { AOM_CDF4(8759, 19519, 25210) },
                { AOM_CDF4(4575, 12767, 17980) },
                { AOM_CDF4(2621, 7677, 12639) },
                { AOM_CDF4(27714, 32397, 32714) },
                { AOM_CDF4(18103, 30373, 32299) },
                { AOM_CDF4(12109, 25700, 29566) },
                { AOM_CDF4(7079, 18230, 24357) },
                { AOM_CDF4(3549, 10632, 15972) },
                { AOM_CDF4(28603, 32338, 32573) },
                { AOM_CDF4(14501, 27887, 30936) },
                { AOM_CDF4(7303, 18808, 25077) },
                { AOM_CDF4(6311, 14321, 20146) },
                { AOM_CDF4(3523, 8456, 15151) },
                { AOM_CDF4(28690, 32538, 32731) },
                { AOM_CDF4(18322, 30307, 32174) },
                { AOM_CDF4(12254, 24410, 29125) },
                { AOM_CDF4(6788, 16946, 24027) },
                { AOM_CDF4(3495, 10962, 17834) },
                { AOM_CDF4(30608, 32739, 32748) },
                { AOM_CDF4(21866, 31864, 32610) },
                { AOM_CDF4(14076, 27533, 31249) },
                { AOM_CDF4(7955, 20570, 27356) },
                { AOM_CDF4(5068, 14634, 21751) },
                { AOM_CDF4(25372, 31738, 32487) },
                { AOM_CDF4(15534, 31068, 32150) },
                { AOM_CDF4(7123, 14247, 18521) },
                { AOM_CDF4(6554, 13107, 26214) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(21845, 28399, 30583) },
                { AOM_CDF4(6951, 29789, 31775) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(10923, 21845, 27307) },
                { AOM_CDF4(13107, 19661, 26214) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(4381, 10541, 14753) },
                { AOM_CDF4(22048, 30479, 32313) },
                { AOM_CDF4(11535, 24774, 29392) },
                { AOM_CDF4(5954, 16992, 24035) },
                { AOM_CDF4(4190, 11244, 17334) },
                { AOM_CDF4(1780, 5000, 8139) },
                { AOM_CDF4(25953, 32034, 32608) },
                { AOM_CDF4(14899, 28326, 31541) },
                { AOM_CDF4(8476, 20965, 27549) },
                { AOM_CDF4(5336, 15381, 21296) },
                { AOM_CDF4(2652, 7064, 11298) },
                { AOM_CDF4(29239, 32295, 32565) },
                { AOM_CDF4(16643, 29322, 31811) },
                { AOM_CDF4(8982, 21640, 27634) },
                { AOM_CDF4(4959, 14646, 21934) },
                { AOM_CDF4(3015, 8316, 13220) },
                { AOM_CDF4(28288, 32518, 32713) },
                { AOM_CDF4(17747, 30236, 32184) },
                { AOM_CDF4(10295, 23995, 29345) },
                { AOM_CDF4(5826, 16941, 24349) },
                { AOM_CDF4(3178, 9120, 13786) },
                { AOM_CDF4(31902, 32735, 32748) },
                { AOM_CDF4(22696, 32042, 32632) },
                { AOM_CDF4(14589, 28476, 31626) },
                { AOM_CDF4(8944, 22030, 28120) },
                { AOM_CDF4(3956, 11107, 16617) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(9365, 18700, 24102) },
                { AOM_CDF4(27453, 32257, 32677) },
                { AOM_CDF4(13479, 27237, 31141) },
                { AOM_CDF4(6703, 18432, 25879) },
                { AOM_CDF4(4428, 11956, 16827) },
                { AOM_CDF4(1202, 4810, 7816) },
                { AOM_CDF4(31162, 32680, 32748) },
                { AOM_CDF4(17116, 29750, 31935) },
                { AOM_CDF4(8416, 19189, 26596) },
                { AOM_CDF4(4723, 14760, 22141) },
                { AOM_CDF4(2979, 8440, 11419) },
                { AOM_CDF4(31523, 32642, 32739) },
                { AOM_CDF4(15899, 29653, 31635) },
                { AOM_CDF4(9674, 20285, 27151) },
                { AOM_CDF4(2427, 14564, 20632) },
                { AOM_CDF4(7282, 12743, 14564) },
                { AOM_CDF4(31554, 32697, 32748) },
                { AOM_CDF4(19940, 30641, 32260) },
                { AOM_CDF4(13598, 26304, 30539) },
                { AOM_CDF4(8791, 18382, 24776) },
                { AOM_CDF4(6036, 9485, 13797) },
                { AOM_CDF4(32588, 32752, 32756) },
                { AOM_CDF4(26038, 32007, 32620) },
                { AOM_CDF4(18455, 29540, 31920) },
                { AOM_CDF4(11320, 24239, 30165) },
                { AOM_CDF4(5860, 14076, 20820) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(2558, 8044, 11808) },
                { AOM_CDF4(19701, 26697, 29994) },
                { AOM_CDF4(9122, 21302, 25761) },
                { AOM_CDF4(4540, 15208, 22319) },
                { AOM_CDF4(3322, 12569, 17865) },
                { AOM_CDF4(1497, 5827, 8874) },
                { AOM_CDF4(26170, 32049, 32618) },
                { AOM_CDF4(14714, 28245, 31747) },
                { AOM_CDF4(8272, 21275, 27141) },
                { AOM_CDF4(5157, 14921, 21979) },
                { AOM_CDF4(2593, 8179, 12749) },
                { AOM_CDF4(28442, 31864, 32496) },
                { AOM_CDF4(16277, 28673, 31455) },
                { AOM_CDF4(8155, 19751, 26384) },
                { AOM_CDF4(4672, 13033, 19120) },
                { AOM_CDF4(2567, 7324, 11401) },
                { AOM_CDF4(28647, 32489, 32652) },
                { AOM_CDF4(18863, 30516, 32256) },
                { AOM_CDF4(10427, 24316, 29768) },
                { AOM_CDF4(6618, 16838, 24042) },
                { AOM_CDF4(3122, 9531, 15133) },
                { AOM_CDF4(32108, 32745, 32749) },
                { AOM_CDF4(23030, 31721, 32601) },
                { AOM_CDF4(12535, 26645, 31076) },
                { AOM_CDF4(8057, 19515, 26907) },
                { AOM_CDF4(4710, 12541, 18985) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } } },
          { { { { AOM_CDF4(5036, 14072, 20572) },
                { AOM_CDF4(22265, 32059, 32670) },
                { AOM_CDF4(12492, 27733, 31751) },
                { AOM_CDF4(7692, 20093, 27379) },
                { AOM_CDF4(5076, 13988, 21755) },
                { AOM_CDF4(3339, 9532, 15261) },
                { AOM_CDF4(24013, 32209, 32664) },
                { AOM_CDF4(14322, 29090, 32163) },
                { AOM_CDF4(7949, 21472, 28640) },
                { AOM_CDF4(6029, 15521, 24286) },
                { AOM_CDF4(3385, 11000, 18762) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(23331, 32119, 32647) },
                { AOM_CDF4(16350, 29938, 32226) },
                { AOM_CDF4(9718, 22205, 29341) },
                { AOM_CDF4(5140, 12208, 23773) },
                { AOM_CDF4(12288, 16384, 20480) },
                { AOM_CDF4(25623, 31888, 32409) },
                { AOM_CDF4(14277, 27629, 31058) },
                { AOM_CDF4(8054, 18551, 25616) },
                { AOM_CDF4(2504, 9008, 16028) },
                { AOM_CDF4(1466, 4912, 9017) },
                { AOM_CDF4(28545, 32401, 32635) },
                { AOM_CDF4(17605, 28557, 31649) },
                { AOM_CDF4(10505, 21155, 26921) },
                { AOM_CDF4(5087, 14196, 20465) },
                { AOM_CDF4(1890, 10082, 16384) },
                { AOM_CDF4(30941, 32581, 32697) },
                { AOM_CDF4(20721, 29919, 32157) },
                { AOM_CDF4(9035, 21239, 27576) },
                { AOM_CDF4(2427, 10923, 15777) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(7947, 20544, 26943) },
                { AOM_CDF4(24652, 32335, 32654) },
                { AOM_CDF4(14826, 29854, 32210) },
                { AOM_CDF4(8809, 22288, 28971) },
                { AOM_CDF4(5767, 15204, 23069) },
                { AOM_CDF4(1365, 6827, 17749) },
                { AOM_CDF4(27298, 32534, 32743) },
                { AOM_CDF4(17987, 31415, 32629) },
                { AOM_CDF4(12542, 25536, 30960) },
                { AOM_CDF4(2731, 16384, 27307) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(27922, 32576, 32720) },
                { AOM_CDF4(24881, 32435, 32657) },
                { AOM_CDF4(6554, 19661, 26214) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(28613, 32174, 32649) },
                { AOM_CDF4(16794, 28570, 32154) },
                { AOM_CDF4(12288, 20480, 26624) },
                { AOM_CDF4(16384, 21845, 27307) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(24030, 31130, 32222) },
                { AOM_CDF4(20194, 30101, 32006) },
                { AOM_CDF4(4681, 18725, 23406) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(22528, 28672, 30720) },
                { AOM_CDF4(21845, 29127, 30948) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(5773, 14964, 20566) },
                { AOM_CDF4(23520, 31988, 32666) },
                { AOM_CDF4(14718, 28826, 31986) },
                { AOM_CDF4(9050, 22915, 28999) },
                { AOM_CDF4(6444, 16812, 23966) },
                { AOM_CDF4(3596, 10705, 16731) },
                { AOM_CDF4(25940, 32530, 32730) },
                { AOM_CDF4(16613, 30412, 32406) },
                { AOM_CDF4(10151, 24748, 30342) },
                { AOM_CDF4(6936, 18162, 25694) },
                { AOM_CDF4(4177, 12450, 19478) },
                { AOM_CDF4(25578, 32184, 32670) },
                { AOM_CDF4(15022, 28821, 31899) },
                { AOM_CDF4(8460, 21731, 28482) },
                { AOM_CDF4(5955, 15692, 22628) },
                { AOM_CDF4(3501, 10096, 15945) },
                { AOM_CDF4(25088, 32313, 32680) },
                { AOM_CDF4(15564, 29262, 32026) },
                { AOM_CDF4(9045, 22016, 28644) },
                { AOM_CDF4(6182, 16554, 23714) },
                { AOM_CDF4(3871, 11212, 17086) },
                { AOM_CDF4(28257, 32655, 32746) },
                { AOM_CDF4(18813, 31355, 32579) },
                { AOM_CDF4(11383, 26157, 31207) },
                { AOM_CDF4(7845, 20824, 28078) },
                { AOM_CDF4(5232, 14795, 22747) },
                { AOM_CDF4(28556, 32271, 32604) },
                { AOM_CDF4(15888, 29004, 31684) },
                { AOM_CDF4(8252, 20004, 26557) },
                { AOM_CDF4(4242, 13050, 20055) },
                { AOM_CDF4(1897, 6457, 11138) },
                { AOM_CDF4(30420, 32590, 32725) },
                { AOM_CDF4(18867, 30189, 32183) },
                { AOM_CDF4(10567, 23955, 29534) },
                { AOM_CDF4(6295, 17592, 25180) },
                { AOM_CDF4(4707, 12220, 18873) },
                { AOM_CDF4(31372, 32677, 32748) },
                { AOM_CDF4(20156, 30965, 32445) },
                { AOM_CDF4(10759, 24575, 30202) },
                { AOM_CDF4(6213, 18391, 25613) },
                { AOM_CDF4(5260, 13498, 20294) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(7910, 18394, 24609) },
                { AOM_CDF4(25137, 32189, 32696) },
                { AOM_CDF4(15399, 29633, 32019) },
                { AOM_CDF4(11449, 23688, 29807) },
                { AOM_CDF4(3171, 16913, 23255) },
                { AOM_CDF4(9638, 11565, 21203) },
                { AOM_CDF4(28689, 32605, 32698) },
                { AOM_CDF4(18748, 31201, 32380) },
                { AOM_CDF4(9854, 26594, 31106) },
                { AOM_CDF4(6275, 18127, 25099) },
                { AOM_CDF4(3641, 10923, 23666) },
                { AOM_CDF4(28173, 32548, 32727) },
                { AOM_CDF4(17401, 30330, 32127) },
                { AOM_CDF4(12663, 25228, 29868) },
                { AOM_CDF4(12208, 25058, 28913) },
                { AOM_CDF4(7123, 19946, 25645) },
                { AOM_CDF4(28164, 32525, 32717) },
                { AOM_CDF4(17988, 30216, 32294) },
                { AOM_CDF4(11252, 24483, 30295) },
                { AOM_CDF4(6199, 15941, 23912) },
                { AOM_CDF4(4096, 16384, 24576) },
                { AOM_CDF4(30688, 32728, 32748) },
                { AOM_CDF4(21740, 32239, 32715) },
                { AOM_CDF4(16132, 26719, 30752) },
                { AOM_CDF4(7282, 18204, 25486) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(27982, 31295, 32400) },
                { AOM_CDF4(16636, 30499, 32012) },
                { AOM_CDF4(4681, 9362, 18725) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(18725, 28087, 30427) },
                { AOM_CDF4(17644, 27727, 30247) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(16384, 21845, 27307) },
                { AOM_CDF4(13107, 19661, 26214) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(6447, 15470, 20891) },
                { AOM_CDF4(23739, 31792, 32623) },
                { AOM_CDF4(13987, 27588, 31524) },
                { AOM_CDF4(7628, 20267, 26781) },
                { AOM_CDF4(5218, 14836, 21280) },
                { AOM_CDF4(2506, 7741, 13200) },
                { AOM_CDF4(27828, 32533, 32744) },
                { AOM_CDF4(17675, 30580, 32387) },
                { AOM_CDF4(10378, 24402, 30056) },
                { AOM_CDF4(7110, 18752, 25433) },
                { AOM_CDF4(4204, 11816, 17754) },
                { AOM_CDF4(29255, 32596, 32746) },
                { AOM_CDF4(17208, 30043, 32288) },
                { AOM_CDF4(10079, 23767, 29801) },
                { AOM_CDF4(6547, 17026, 24483) },
                { AOM_CDF4(3610, 11158, 16761) },
                { AOM_CDF4(28656, 32539, 32743) },
                { AOM_CDF4(17009, 30249, 32364) },
                { AOM_CDF4(10196, 23531, 29486) },
                { AOM_CDF4(6325, 17030, 24183) },
                { AOM_CDF4(3724, 10255, 16668) },
                { AOM_CDF4(30777, 32733, 32748) },
                { AOM_CDF4(21665, 32002, 32651) },
                { AOM_CDF4(13584, 28143, 31783) },
                { AOM_CDF4(8635, 22029, 28686) },
                { AOM_CDF4(5640, 16027, 23429) },
                { AOM_CDF4(29265, 32394, 32668) },
                { AOM_CDF4(15949, 28837, 31668) },
                { AOM_CDF4(8609, 20421, 26119) },
                { AOM_CDF4(4078, 11724, 17986) },
                { AOM_CDF4(2264, 7149, 12035) },
                { AOM_CDF4(31061, 32665, 32737) },
                { AOM_CDF4(21073, 31231, 32395) },
                { AOM_CDF4(12636, 25743, 29942) },
                { AOM_CDF4(6090, 17544, 25953) },
                { AOM_CDF4(3855, 11290, 18449) },
                { AOM_CDF4(31680, 32745, 32749) },
                { AOM_CDF4(22609, 31806, 32593) },
                { AOM_CDF4(13415, 26899, 31068) },
                { AOM_CDF4(5221, 20216, 25215) },
                { AOM_CDF4(5041, 14564, 22125) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(5079, 14099, 20130) },
                { AOM_CDF4(24576, 31177, 32414) },
                { AOM_CDF4(11252, 26584, 30406) },
                { AOM_CDF4(7509, 17067, 26169) },
                { AOM_CDF4(5334, 14479, 22099) },
                { AOM_CDF4(6242, 12483, 18725) },
                { AOM_CDF4(28608, 32613, 32716) },
                { AOM_CDF4(15918, 30477, 32535) },
                { AOM_CDF4(7232, 23051, 30734) },
                { AOM_CDF4(5120, 14336, 25600) },
                { AOM_CDF4(5958, 14895, 20852) },
                { AOM_CDF4(29156, 32252, 32621) },
                { AOM_CDF4(15560, 28153, 31449) },
                { AOM_CDF4(9242, 22686, 27727) },
                { AOM_CDF4(4681, 14043, 23406) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(28863, 32306, 32697) },
                { AOM_CDF4(15547, 29563, 31811) },
                { AOM_CDF4(9125, 21154, 26961) },
                { AOM_CDF4(5699, 14247, 21370) },
                { AOM_CDF4(6554, 19661, 26214) },
                { AOM_CDF4(32103, 32748, 32752) },
                { AOM_CDF4(23758, 32223, 32744) },
                { AOM_CDF4(16193, 30101, 32577) },
                { AOM_CDF4(14564, 27307, 30948) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(18432, 28672, 30720) },
                { AOM_CDF4(10128, 30385, 32172) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(5261, 13528, 19157) },
                { AOM_CDF4(21845, 30866, 32344) },
                { AOM_CDF4(13118, 26804, 31002) },
                { AOM_CDF4(6350, 18929, 25715) },
                { AOM_CDF4(5461, 14321, 20814) },
                { AOM_CDF4(2306, 7336, 12017) },
                { AOM_CDF4(27196, 32417, 32741) },
                { AOM_CDF4(15830, 29718, 32103) },
                { AOM_CDF4(10241, 23629, 29542) },
                { AOM_CDF4(6312, 17289, 24445) },
                { AOM_CDF4(2833, 9264, 15465) },
                { AOM_CDF4(29703, 32522, 32729) },
                { AOM_CDF4(16095, 29765, 32162) },
                { AOM_CDF4(9780, 22879, 28430) },
                { AOM_CDF4(5567, 16129, 23510) },
                { AOM_CDF4(3251, 9383, 14999) },
                { AOM_CDF4(28383, 32545, 32719) },
                { AOM_CDF4(16464, 30234, 32335) },
                { AOM_CDF4(10502, 23622, 29627) },
                { AOM_CDF4(5559, 17779, 24441) },
                { AOM_CDF4(3348, 10162, 15999) },
                { AOM_CDF4(32165, 32746, 32750) },
                { AOM_CDF4(22468, 31927, 32605) },
                { AOM_CDF4(13921, 27871, 31412) },
                { AOM_CDF4(8829, 21970, 28148) },
                { AOM_CDF4(5233, 14474, 21191) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(6851, 15882, 21938) },
                { AOM_CDF4(26139, 31632, 32484) },
                { AOM_CDF4(12047, 25178, 30720) },
                { AOM_CDF4(3091, 11129, 19784) },
                { AOM_CDF4(2657, 7971, 14170) },
                { AOM_CDF4(3745, 5617, 7490) },
                { AOM_CDF4(29314, 32307, 32735) },
                { AOM_CDF4(15065, 28741, 31796) },
                { AOM_CDF4(8764, 20194, 26291) },
                { AOM_CDF4(3390, 18079, 24858) },
                { AOM_CDF4(1214, 10923, 18204) },
                { AOM_CDF4(29383, 32257, 32576) },
                { AOM_CDF4(15951, 29306, 31531) },
                { AOM_CDF4(9466, 18204, 24758) },
                { AOM_CDF4(7282, 18204, 21845) },
                { AOM_CDF4(5461, 16384, 21845) },
                { AOM_CDF4(30623, 32546, 32694) },
                { AOM_CDF4(15882, 29424, 31765) },
                { AOM_CDF4(7944, 20852, 28796) },
                { AOM_CDF4(4681, 9362, 14043) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(32611, 32752, 32756) },
                { AOM_CDF4(25407, 31995, 32711) },
                { AOM_CDF4(17172, 31035, 32138) },
                { AOM_CDF4(1820, 19115, 27307) },
                { AOM_CDF4(10240, 18432, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } },
            { { { AOM_CDF4(3604, 9595, 14487) },
                { AOM_CDF4(22240, 30555, 32310) },
                { AOM_CDF4(11403, 24475, 29405) },
                { AOM_CDF4(6591, 17845, 24063) },
                { AOM_CDF4(3627, 12092, 19467) },
                { AOM_CDF4(3116, 6936, 10956) },
                { AOM_CDF4(27185, 32314, 32707) },
                { AOM_CDF4(15925, 29207, 32059) },
                { AOM_CDF4(9406, 22595, 29036) },
                { AOM_CDF4(5099, 14736, 22260) },
                { AOM_CDF4(3579, 9613, 14946) },
                { AOM_CDF4(29825, 32513, 32687) },
                { AOM_CDF4(15466, 29317, 31843) },
                { AOM_CDF4(8931, 22121, 27918) },
                { AOM_CDF4(5323, 16800, 23453) },
                { AOM_CDF4(2782, 9274, 13602) },
                { AOM_CDF4(28480, 32459, 32680) },
                { AOM_CDF4(17344, 30347, 32335) },
                { AOM_CDF4(9916, 23651, 29660) },
                { AOM_CDF4(6376, 16606, 24835) },
                { AOM_CDF4(2988, 13147, 18525) },
                { AOM_CDF4(32179, 32745, 32749) },
                { AOM_CDF4(21875, 31678, 32604) },
                { AOM_CDF4(13014, 27171, 31206) },
                { AOM_CDF4(8696, 20879, 27937) },
                { AOM_CDF4(5451, 14185, 20751) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } },
              { { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) },
                { AOM_CDF4(8192, 16384, 24576) } } } },
        };

static const aom_cdf_prob av1_default_coeff_base_eob_multi_cdfs
    [TOKEN_CDF_Q_CTXS][TX_SIZES][PLANE_TYPES][SIG_COEF_CONTEXTS_EOB]
    [CDF_SIZE(NUM_BASE_LEVELS + 1)] = {
      { { { { AOM_CDF3(20597, 31183) },
            { AOM_CDF3(31273, 32627) },
            { AOM_CDF3(31825, 32680) },
            { AOM_CDF3(27224, 30407) } },
          { { AOM_CDF3(28394, 31807) },
            { AOM_CDF3(31660, 32638) },
            { AOM_CDF3(32058, 32667) },
            { AOM_CDF3(31564, 32518) } } },
        { { { AOM_CDF3(15187, 30433) },
            { AOM_CDF3(32209, 32689) },
            { AOM_CDF3(32628, 32745) },
            { AOM_CDF3(29289, 32300) } },
          { { AOM_CDF3(26024, 31469) },
            { AOM_CDF3(31866, 32605) },
            { AOM_CDF3(32328, 32726) },
            { AOM_CDF3(32366, 32703) } } },
        { { { AOM_CDF3(5796, 25887) },
            { AOM_CDF3(32280, 32684) },
            { AOM_CDF3(32718, 32756) },
            { AOM_CDF3(32341, 32689) } },
          { { AOM_CDF3(17040, 25504) },
            { AOM_CDF3(32367, 32732) },
            { AOM_CDF3(32104, 32702) },
            { AOM_CDF3(32536, 32731) } } },
        { { { AOM_CDF3(10486, 20733) },
            { AOM_CDF3(31320, 32449) },
            { AOM_CDF3(32303, 32722) },
            { AOM_CDF3(32517, 32704) } },
          { { AOM_CDF3(27125, 29437) },
            { AOM_CDF3(31058, 31898) },
            { AOM_CDF3(31343, 32056) },
            { AOM_CDF3(31555, 32742) } } },
        { { { AOM_CDF3(16384, 19661) },
            { AOM_CDF3(32276, 32589) },
            { AOM_CDF3(30554, 32547) },
            { AOM_CDF3(32549, 32713) } },
          { { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) } } } },
      { { { { AOM_CDF3(20771, 31640) },
            { AOM_CDF3(31727, 32593) },
            { AOM_CDF3(31842, 32688) },
            { AOM_CDF3(29451, 31534) } },
          { { AOM_CDF3(27842, 31527) },
            { AOM_CDF3(31658, 32647) },
            { AOM_CDF3(32108, 32699) },
            { AOM_CDF3(32038, 32662) } } },
        { { { AOM_CDF3(17666, 30666) },
            { AOM_CDF3(32305, 32679) },
            { AOM_CDF3(32591, 32751) },
            { AOM_CDF3(31284, 32594) } },
          { { AOM_CDF3(28021, 31540) },
            { AOM_CDF3(31750, 32595) },
            { AOM_CDF3(32403, 32725) },
            { AOM_CDF3(32425, 32722) } } },
        { { { AOM_CDF3(12056, 29503) },
            { AOM_CDF3(32342, 32686) },
            { AOM_CDF3(32719, 32756) },
            { AOM_CDF3(32590, 32738) } },
          { { AOM_CDF3(21009, 28705) },
            { AOM_CDF3(31851, 32617) },
            { AOM_CDF3(32262, 32743) },
            { AOM_CDF3(32547, 32721) } } },
        { { { AOM_CDF3(3890, 15934) },
            { AOM_CDF3(30573, 31743) },
            { AOM_CDF3(32509, 32749) },
            { AOM_CDF3(32625, 32744) } },
          { { AOM_CDF3(22429, 27807) },
            { AOM_CDF3(30492, 31850) },
            { AOM_CDF3(31374, 32419) },
            { AOM_CDF3(31007, 32717) } } },
        { { { AOM_CDF3(4830, 5875) },
            { AOM_CDF3(29974, 31257) },
            { AOM_CDF3(31928, 32628) },
            { AOM_CDF3(31925, 32734) } },
          { { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) } } } },
      { { { { AOM_CDF3(22884, 31609) },
            { AOM_CDF3(31723, 32529) },
            { AOM_CDF3(31905, 32699) },
            { AOM_CDF3(30839, 32165) } },
          { { AOM_CDF3(27149, 31080) },
            { AOM_CDF3(31720, 32693) },
            { AOM_CDF3(32371, 32721) },
            { AOM_CDF3(32387, 32743) } } },
        { { { AOM_CDF3(21572, 31546) },
            { AOM_CDF3(32053, 32643) },
            { AOM_CDF3(32497, 32733) },
            { AOM_CDF3(32270, 32655) } },
          { { AOM_CDF3(27952, 31297) },
            { AOM_CDF3(31833, 32635) },
            { AOM_CDF3(32444, 32747) },
            { AOM_CDF3(32513, 32742) } } },
        { { { AOM_CDF3(16700, 29042) },
            { AOM_CDF3(32185, 32665) },
            { AOM_CDF3(32667, 32740) },
            { AOM_CDF3(32619, 32747) } },
          { { AOM_CDF3(25356, 30314) },
            { AOM_CDF3(31958, 32606) },
            { AOM_CDF3(32451, 32715) },
            { AOM_CDF3(32602, 32747) } } },
        { { { AOM_CDF3(12218, 22903) },
            { AOM_CDF3(31421, 32164) },
            { AOM_CDF3(32454, 32748) },
            { AOM_CDF3(32174, 32570) } },
          { { AOM_CDF3(22272, 28451) },
            { AOM_CDF3(30802, 32293) },
            { AOM_CDF3(30093, 32099) },
            { AOM_CDF3(30899, 32725) } } },
        { { { AOM_CDF3(5670, 16547) },
            { AOM_CDF3(28947, 30241) },
            { AOM_CDF3(32105, 32635) },
            { AOM_CDF3(32316, 32693) } },
          { { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) } } } },
      { { { { AOM_CDF3(23556, 31195) },
            { AOM_CDF3(31953, 32674) },
            { AOM_CDF3(32129, 32703) },
            { AOM_CDF3(31963, 32607) } },
          { { AOM_CDF3(27624, 31497) },
            { AOM_CDF3(32234, 32756) },
            { AOM_CDF3(32558, 32740) },
            { AOM_CDF3(32640, 32747) } } },
        { { { AOM_CDF3(23114, 31301) },
            { AOM_CDF3(32277, 32729) },
            { AOM_CDF3(32435, 32723) },
            { AOM_CDF3(32395, 32706) } },
          { { AOM_CDF3(27864, 31264) },
            { AOM_CDF3(31905, 32658) },
            { AOM_CDF3(32506, 32731) },
            { AOM_CDF3(32618, 32730) } } },
        { { { AOM_CDF3(21525, 30206) },
            { AOM_CDF3(32186, 32691) },
            { AOM_CDF3(32572, 32724) },
            { AOM_CDF3(32656, 32744) } },
          { { AOM_CDF3(27055, 30657) },
            { AOM_CDF3(32025, 32586) },
            { AOM_CDF3(31279, 32272) },
            { AOM_CDF3(31858, 32313) } } },
        { { { AOM_CDF3(21773, 28489) },
            { AOM_CDF3(32495, 32732) },
            { AOM_CDF3(32332, 32728) },
            { AOM_CDF3(30806, 32184) } },
          { { AOM_CDF3(27061, 30422) },
            { AOM_CDF3(32361, 32731) },
            { AOM_CDF3(21845, 27307) },
            { AOM_CDF3(31694, 32231) } } },
        { { { AOM_CDF3(12922, 23047) },
            { AOM_CDF3(31948, 32657) },
            { AOM_CDF3(32165, 32567) },
            { AOM_CDF3(31591, 32376) } },
          { { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) },
            { AOM_CDF3(10923, 21845) } } } },
    };

static const aom_cdf_prob av1_default_eob_extra_cdfs
    [TOKEN_CDF_Q_CTXS][TX_SIZES][PLANE_TYPES][EOB_COEF_CONTEXTS][CDF_SIZE(2)] =
        {
          { { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(14583) }, { AOM_CDF2(14609) }, { AOM_CDF2(6039) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(18056) }, { AOM_CDF2(16946) }, { AOM_CDF2(15356) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16960) }, { AOM_CDF2(18623) }, { AOM_CDF2(17045) },
                { AOM_CDF2(11983) }, { AOM_CDF2(9010) },  { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(21542) }, { AOM_CDF2(20073) }, { AOM_CDF2(16775) },
                { AOM_CDF2(20287) }, { AOM_CDF2(21990) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(21749) }, { AOM_CDF2(18940) }, { AOM_CDF2(16287) },
                { AOM_CDF2(18244) }, { AOM_CDF2(14240) }, { AOM_CDF2(15856) },
                { AOM_CDF2(16732) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(22957) }, { AOM_CDF2(18467) }, { AOM_CDF2(18271) },
                { AOM_CDF2(18538) }, { AOM_CDF2(13782) }, { AOM_CDF2(23742) },
                { AOM_CDF2(20798) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(20661) }, { AOM_CDF2(22431) }, { AOM_CDF2(22648) },
                { AOM_CDF2(20514) }, { AOM_CDF2(19787) }, { AOM_CDF2(13765) },
                { AOM_CDF2(19496) }, { AOM_CDF2(11160) }, { AOM_CDF2(11234) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(25541) }, { AOM_CDF2(25051) }, { AOM_CDF2(21781) },
                { AOM_CDF2(20439) }, { AOM_CDF2(21370) }, { AOM_CDF2(18569) },
                { AOM_CDF2(13370) }, { AOM_CDF2(27280) }, { AOM_CDF2(27699) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(19181) }, { AOM_CDF2(16733) }, { AOM_CDF2(18422) },
                { AOM_CDF2(15124) }, { AOM_CDF2(22686) }, { AOM_CDF2(20549) },
                { AOM_CDF2(14489) }, { AOM_CDF2(15881) }, { AOM_CDF2(18639) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } } },
          { { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(14911) }, { AOM_CDF2(15414) }, { AOM_CDF2(9035) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(19376) }, { AOM_CDF2(18353) }, { AOM_CDF2(19497) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(17395) }, { AOM_CDF2(18844) }, { AOM_CDF2(17566) },
                { AOM_CDF2(14742) }, { AOM_CDF2(10578) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(20738) }, { AOM_CDF2(20524) }, { AOM_CDF2(17537) },
                { AOM_CDF2(24074) }, { AOM_CDF2(26643) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(19326) }, { AOM_CDF2(18988) }, { AOM_CDF2(16704) },
                { AOM_CDF2(19532) }, { AOM_CDF2(15510) }, { AOM_CDF2(19016) },
                { AOM_CDF2(20232) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(22906) }, { AOM_CDF2(22105) }, { AOM_CDF2(21782) },
                { AOM_CDF2(20700) }, { AOM_CDF2(14921) }, { AOM_CDF2(23591) },
                { AOM_CDF2(24978) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(25305) }, { AOM_CDF2(23310) }, { AOM_CDF2(18178) },
                { AOM_CDF2(19735) }, { AOM_CDF2(20520) }, { AOM_CDF2(15769) },
                { AOM_CDF2(23096) }, { AOM_CDF2(13287) }, { AOM_CDF2(12667) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(26633) }, { AOM_CDF2(22458) }, { AOM_CDF2(22068) },
                { AOM_CDF2(19405) }, { AOM_CDF2(23887) }, { AOM_CDF2(8856) },
                { AOM_CDF2(16328) }, { AOM_CDF2(31572) }, { AOM_CDF2(26214) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(26354) }, { AOM_CDF2(20723) }, { AOM_CDF2(20647) },
                { AOM_CDF2(15541) }, { AOM_CDF2(21027) }, { AOM_CDF2(22805) },
                { AOM_CDF2(12712) }, { AOM_CDF2(17246) }, { AOM_CDF2(21475) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } } },
          { { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(15912) }, { AOM_CDF2(16777) }, { AOM_CDF2(13160) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(19931) }, { AOM_CDF2(19777) }, { AOM_CDF2(22966) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(18112) }, { AOM_CDF2(19084) }, { AOM_CDF2(18295) },
                { AOM_CDF2(16862) }, { AOM_CDF2(15713) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(21534) }, { AOM_CDF2(20739) }, { AOM_CDF2(19526) },
                { AOM_CDF2(22958) }, { AOM_CDF2(28379) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(17715) }, { AOM_CDF2(18575) }, { AOM_CDF2(17600) },
                { AOM_CDF2(19592) }, { AOM_CDF2(17971) }, { AOM_CDF2(19944) },
                { AOM_CDF2(19606) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(23762) }, { AOM_CDF2(25444) }, { AOM_CDF2(23182) },
                { AOM_CDF2(21878) }, { AOM_CDF2(13577) }, { AOM_CDF2(21873) },
                { AOM_CDF2(29739) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(25025) }, { AOM_CDF2(22643) }, { AOM_CDF2(17867) },
                { AOM_CDF2(20472) }, { AOM_CDF2(22017) }, { AOM_CDF2(19198) },
                { AOM_CDF2(24447) }, { AOM_CDF2(15542) }, { AOM_CDF2(18678) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(25073) }, { AOM_CDF2(23406) }, { AOM_CDF2(21609) },
                { AOM_CDF2(22615) }, { AOM_CDF2(23636) }, { AOM_CDF2(6827) },
                { AOM_CDF2(20312) }, { AOM_CDF2(31618) }, { AOM_CDF2(21845) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(27530) }, { AOM_CDF2(20747) }, { AOM_CDF2(21651) },
                { AOM_CDF2(18812) }, { AOM_CDF2(17457) }, { AOM_CDF2(22176) },
                { AOM_CDF2(12059) }, { AOM_CDF2(18204) }, { AOM_CDF2(20602) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } } },
          { { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(15381) }, { AOM_CDF2(17118) }, { AOM_CDF2(18874) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(20759) }, { AOM_CDF2(23346) }, { AOM_CDF2(26502) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(17675) }, { AOM_CDF2(19214) }, { AOM_CDF2(18992) },
                { AOM_CDF2(20161) }, { AOM_CDF2(21785) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(24438) }, { AOM_CDF2(25137) }, { AOM_CDF2(24678) },
                { AOM_CDF2(24700) }, { AOM_CDF2(20852) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(20315) }, { AOM_CDF2(18453) }, { AOM_CDF2(18924) },
                { AOM_CDF2(20072) }, { AOM_CDF2(19729) }, { AOM_CDF2(23392) },
                { AOM_CDF2(25556) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(24534) }, { AOM_CDF2(28135) }, { AOM_CDF2(22390) },
                { AOM_CDF2(21958) }, { AOM_CDF2(18725) }, { AOM_CDF2(24273) },
                { AOM_CDF2(31403) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(22457) }, { AOM_CDF2(23403) }, { AOM_CDF2(19857) },
                { AOM_CDF2(20241) }, { AOM_CDF2(22129) }, { AOM_CDF2(20539) },
                { AOM_CDF2(20621) }, { AOM_CDF2(21215) }, { AOM_CDF2(26935) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(24305) }, { AOM_CDF2(24004) }, { AOM_CDF2(20950) },
                { AOM_CDF2(25486) }, { AOM_CDF2(18971) }, { AOM_CDF2(23406) },
                { AOM_CDF2(17873) }, { AOM_CDF2(30840) }, { AOM_CDF2(10923) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } },
            { { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(25699) }, { AOM_CDF2(21787) }, { AOM_CDF2(21964) },
                { AOM_CDF2(17686) }, { AOM_CDF2(21433) }, { AOM_CDF2(16384) },
                { AOM_CDF2(18530) }, { AOM_CDF2(21633) }, { AOM_CDF2(14398) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } },
              { { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) }, { AOM_CDF2(16384) }, { AOM_CDF2(16384) },
                { AOM_CDF2(16384) } } } },
        };

static const aom_cdf_prob av1_default_coeff_lps_multi_cdfs
    [TOKEN_CDF_Q_CTXS][TX_SIZES][PLANE_TYPES][LEVEL_CONTEXTS]
    [CDF_SIZE(BR_CDF_SIZE)] = {
      { { { { AOM_CDF4(19551, 26389, 29394) },
            { AOM_CDF4(15086, 22440, 26364) },
            { AOM_CDF4(9867, 16572, 21009) },
            { AOM_CDF4(6574, 12039, 16670) },
            { AOM_CDF4(4817, 9114, 12770) },
            { AOM_CDF4(3666, 7177, 10609) },
            { AOM_CDF4(1746, 3398, 5058) },
            { AOM_CDF4(22294, 28462, 30592) },
            { AOM_CDF4(18791, 26002, 29147) },
            { AOM_CDF4(13982, 21661, 25851) },
            { AOM_CDF4(10125, 17144, 21891) },
            { AOM_CDF4(7478, 13554, 18127) },
            { AOM_CDF4(5838, 10966, 15307) },
            { AOM_CDF4(3097, 5933, 8647) },
            { AOM_CDF4(17134, 23834, 26964) },
            { AOM_CDF4(15992, 23132, 26695) },
            { AOM_CDF4(12546, 19975, 24251) },
            { AOM_CDF4(9750, 16337, 21035) },
            { AOM_CDF4(7560, 13564, 18188) },
            { AOM_CDF4(6154, 11431, 15727) },
            { AOM_CDF4(3699, 7221, 10274) } },
          { { AOM_CDF4(18428, 24992, 28179) },
            { AOM_CDF4(14152, 21273, 25185) },
            { AOM_CDF4(9541, 16148, 20496) },
            { AOM_CDF4(6782, 12221, 16530) },
            { AOM_CDF4(5190, 9734, 13857) },
            { AOM_CDF4(4244, 8233, 11808) },
            { AOM_CDF4(2563, 5107, 7516) },
            { AOM_CDF4(24616, 29743, 31576) },
            { AOM_CDF4(20086, 27180, 30135) },
            { AOM_CDF4(14955, 23030, 27367) },
            { AOM_CDF4(11263, 18700, 23866) },
            { AOM_CDF4(8653, 15468, 20410) },
            { AOM_CDF4(7365, 13249, 18188) },
            { AOM_CDF4(4647, 8900, 12706) },
            { AOM_CDF4(23742, 29468, 31373) },
            { AOM_CDF4(21826, 28503, 30879) },
            { AOM_CDF4(16900, 25099, 28993) },
            { AOM_CDF4(12951, 21209, 25853) },
            { AOM_CDF4(10994, 18168, 23086) },
            { AOM_CDF4(8611, 15481, 20320) },
            { AOM_CDF4(6804, 11905, 16212) } } },
        { { { AOM_CDF4(18147, 24912, 28104) },
            { AOM_CDF4(13981, 21382, 25277) },
            { AOM_CDF4(9783, 16554, 21286) },
            { AOM_CDF4(6894, 12423, 16934) },
            { AOM_CDF4(5308, 9739, 13703) },
            { AOM_CDF4(3879, 7778, 11189) },
            { AOM_CDF4(1754, 3458, 5126) },
            { AOM_CDF4(18387, 24745, 27571) },
            { AOM_CDF4(16925, 24036, 27536) },
            { AOM_CDF4(13403, 20827, 24997) },
            { AOM_CDF4(10016, 16766, 21323) },
            { AOM_CDF4(7571, 13492, 17918) },
            { AOM_CDF4(6033, 11142, 15323) },
            { AOM_CDF4(2946, 5727, 8243) },
            { AOM_CDF4(21853, 27668, 30027) },
            { AOM_CDF4(19575, 26398, 29328) },
            { AOM_CDF4(15913, 23532, 27355) },
            { AOM_CDF4(12062, 19606, 24209) },
            { AOM_CDF4(9277, 16022, 20814) },
            { AOM_CDF4(7565, 13521, 18188) },
            { AOM_CDF4(4145, 7899, 11241) } },
          { { AOM_CDF4(18987, 25320, 28207) },
            { AOM_CDF4(14495, 21416, 24953) },
            { AOM_CDF4(9644, 16089, 20432) },
            { AOM_CDF4(6209, 11539, 15700) },
            { AOM_CDF4(4614, 8897, 12620) },
            { AOM_CDF4(3707, 7058, 10428) },
            { AOM_CDF4(2133, 4370, 6280) },
            { AOM_CDF4(20230, 25658, 28285) },
            { AOM_CDF4(18809, 25538, 28589) },
            { AOM_CDF4(15223, 22386, 26228) },
            { AOM_CDF4(11028, 18262, 22688) },
            { AOM_CDF4(8057, 14435, 19260) },
            { AOM_CDF4(6957, 12514, 16745) },
            { AOM_CDF4(4547, 8769, 12279) },
            { AOM_CDF4(24638, 30008, 31594) },
            { AOM_CDF4(22399, 28808, 31134) },
            { AOM_CDF4(17539, 25444, 29031) },
            { AOM_CDF4(12927, 20996, 25632) },
            { AOM_CDF4(10192, 17374, 22231) },
            { AOM_CDF4(8456, 14966, 20063) },
            { AOM_CDF4(5218, 10245, 14469) } } },
        { { { AOM_CDF4(16392, 22702, 25958) },
            { AOM_CDF4(12223, 18336, 22043) },
            { AOM_CDF4(7223, 12257, 16027) },
            { AOM_CDF4(5147, 9350, 12639) },
            { AOM_CDF4(4001, 7730, 10664) },
            { AOM_CDF4(3231, 6521, 9334) },
            { AOM_CDF4(1401, 2882, 4279) },
            { AOM_CDF4(18215, 24183, 27317) },
            { AOM_CDF4(17044, 24122, 27485) },
            { AOM_CDF4(13672, 21231, 25203) },
            { AOM_CDF4(10558, 17393, 21747) },
            { AOM_CDF4(8040, 13822, 18467) },
            { AOM_CDF4(6203, 11237, 15630) },
            { AOM_CDF4(2705, 5161, 7397) },
            { AOM_CDF4(23884, 28581, 30490) },
            { AOM_CDF4(22647, 28396, 30485) },
            { AOM_CDF4(18533, 25799, 28974) },
            { AOM_CDF4(13947, 21648, 25883) },
            { AOM_CDF4(10384, 17675, 22355) },
            { AOM_CDF4(8315, 14706, 19350) },
            { AOM_CDF4(4020, 7618, 10742) } },
          { { AOM_CDF4(18997, 24785, 27601) },
            { AOM_CDF4(12268, 18352, 21872) },
            { AOM_CDF4(7051, 12368, 15536) },
            { AOM_CDF4(5731, 9979, 13350) },
            { AOM_CDF4(4499, 7630, 10368) },
            { AOM_CDF4(3875, 5757, 8524) },
            { AOM_CDF4(1056, 2589, 3853) },
            { AOM_CDF4(17554, 20285, 23016) },
            { AOM_CDF4(15322, 22897, 26226) },
            { AOM_CDF4(10940, 18004, 21987) },
            { AOM_CDF4(9362, 15739, 19471) },
            { AOM_CDF4(6741, 11947, 16555) },
            { AOM_CDF4(5928, 10209, 15149) },
            { AOM_CDF4(2893, 5825, 8187) },
            { AOM_CDF4(25166, 30849, 31919) },
            { AOM_CDF4(21563, 28560, 30954) },
            { AOM_CDF4(18615, 26000, 29479) },
            { AOM_CDF4(15017, 22815, 26675) },
            { AOM_CDF4(11993, 19544, 23987) },
            { AOM_CDF4(9525, 15728, 20908) },
            { AOM_CDF4(6177, 11184, 15150) } } },
        { { { AOM_CDF4(10837, 15701, 19712) },
            { AOM_CDF4(9385, 14893, 18343) },
            { AOM_CDF4(7579, 13650, 17912) },
            { AOM_CDF4(5988, 11782, 16045) },
            { AOM_CDF4(4751, 8271, 11819) },
            { AOM_CDF4(2707, 6174, 9688) },
            { AOM_CDF4(1305, 2512, 3937) },
            { AOM_CDF4(10791, 22743, 28173) },
            { AOM_CDF4(9839, 17031, 22109) },
            { AOM_CDF4(8967, 15846, 19602) },
            { AOM_CDF4(7062, 12429, 17231) },
            { AOM_CDF4(5855, 10401, 14220) },
            { AOM_CDF4(4695, 8820, 13562) },
            { AOM_CDF4(2180, 4064, 5805) },
            { AOM_CDF4(17254, 21692, 24122) },
            { AOM_CDF4(20856, 26923, 29298) },
            { AOM_CDF4(18254, 24995, 28332) },
            { AOM_CDF4(13954, 21179, 25354) },
            { AOM_CDF4(10291, 17387, 21898) },
            { AOM_CDF4(8179, 14580, 19018) },
            { AOM_CDF4(3337, 6302, 9006) } },
          { { AOM_CDF4(14532, 19661, 23935) },
            { AOM_CDF4(9596, 13107, 15682) },
            { AOM_CDF4(8919, 13573, 16869) },
            { AOM_CDF4(6687, 14043, 18056) },
            { AOM_CDF4(5617, 7490, 14043) },
            { AOM_CDF4(4520, 6780, 11299) },
            { AOM_CDF4(1008, 2521, 3529) },
            { AOM_CDF4(14848, 28160, 31744) },
            { AOM_CDF4(14434, 22235, 28477) },
            { AOM_CDF4(10012, 19115, 23666) },
            { AOM_CDF4(5461, 8582, 13263) },
            { AOM_CDF4(8192, 9830, 13107) },
            { AOM_CDF4(6554, 8738, 13107) },
            { AOM_CDF4(1394, 2789, 5578) },
            { AOM_CDF4(23991, 30427, 32300) },
            { AOM_CDF4(20335, 28072, 31053) },
            { AOM_CDF4(16143, 24723, 28638) },
            { AOM_CDF4(14613, 22703, 27698) },
            { AOM_CDF4(14407, 21936, 25727) },
            { AOM_CDF4(13899, 21999, 26141) },
            { AOM_CDF4(9849, 16897, 23246) } } },
        { { { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } },
          { { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } } } },
      { { { { AOM_CDF4(19774, 26358, 29082) },
            { AOM_CDF4(14793, 22128, 26184) },
            { AOM_CDF4(8992, 15426, 19832) },
            { AOM_CDF4(5843, 10899, 15174) },
            { AOM_CDF4(4555, 8453, 12187) },
            { AOM_CDF4(3314, 6670, 10157) },
            { AOM_CDF4(1952, 3884, 5849) },
            { AOM_CDF4(22545, 28223, 30308) },
            { AOM_CDF4(19259, 26622, 29670) },
            { AOM_CDF4(13873, 21713, 25867) },
            { AOM_CDF4(10038, 17342, 22311) },
            { AOM_CDF4(7772, 13947, 18771) },
            { AOM_CDF4(6253, 11629, 16157) },
            { AOM_CDF4(3854, 7371, 10782) },
            { AOM_CDF4(20735, 27052, 29706) },
            { AOM_CDF4(18100, 25463, 28757) },
            { AOM_CDF4(14127, 22191, 26638) },
            { AOM_CDF4(10740, 18111, 23164) },
            { AOM_CDF4(8260, 15001, 20096) },
            { AOM_CDF4(7063, 13121, 17826) },
            { AOM_CDF4(4926, 9401, 13266) } },
          { { AOM_CDF4(18656, 25450, 28563) },
            { AOM_CDF4(14172, 21601, 25681) },
            { AOM_CDF4(9863, 16732, 21478) },
            { AOM_CDF4(7011, 12934, 17759) },
            { AOM_CDF4(5450, 10634, 14969) },
            { AOM_CDF4(4454, 8494, 12264) },
            { AOM_CDF4(2820, 5975, 8764) },
            { AOM_CDF4(25785, 30620, 32085) },
            { AOM_CDF4(21551, 28552, 31100) },
            { AOM_CDF4(16467, 24836, 28985) },
            { AOM_CDF4(12615, 20070, 25452) },
            { AOM_CDF4(9175, 16469, 21887) },
            { AOM_CDF4(8129, 14184, 19537) },
            { AOM_CDF4(5565, 10857, 15333) },
            { AOM_CDF4(26325, 30916, 31855) },
            { AOM_CDF4(24536, 30416, 31975) },
            { AOM_CDF4(19786, 27515, 30609) },
            { AOM_CDF4(14246, 23341, 27922) },
            { AOM_CDF4(11548, 19018, 24504) },
            { AOM_CDF4(10118, 18051, 22765) },
            { AOM_CDF4(6683, 13150, 19187) } } },
        { { { AOM_CDF4(18898, 25448, 28561) },
            { AOM_CDF4(14653, 22247, 26134) },
            { AOM_CDF4(9800, 16705, 21415) },
            { AOM_CDF4(6649, 12024, 16547) },
            { AOM_CDF4(4799, 9158, 12993) },
            { AOM_CDF4(3652, 7164, 10409) },
            { AOM_CDF4(1788, 3504, 5273) },
            { AOM_CDF4(19949, 26182, 28667) },
            { AOM_CDF4(17762, 24858, 28208) },
            { AOM_CDF4(13405, 20936, 25298) },
            { AOM_CDF4(9720, 16629, 21324) },
            { AOM_CDF4(7426, 13427, 17962) },
            { AOM_CDF4(6115, 11311, 15573) },
            { AOM_CDF4(3493, 6708, 9673) },
            { AOM_CDF4(23508, 29198, 31135) },
            { AOM_CDF4(21007, 27758, 30388) },
            { AOM_CDF4(16514, 24271, 28148) },
            { AOM_CDF4(12383, 20230, 24979) },
            { AOM_CDF4(9679, 16637, 21660) },
            { AOM_CDF4(7994, 14236, 19060) },
            { AOM_CDF4(5020, 9445, 13404) } },
          { { AOM_CDF4(18592, 24875, 28038) },
            { AOM_CDF4(13893, 20779, 24708) },
            { AOM_CDF4(9044, 15352, 19728) },
            { AOM_CDF4(6146, 11585, 15632) },
            { AOM_CDF4(4577, 8600, 12564) },
            { AOM_CDF4(3854, 7249, 10820) },
            { AOM_CDF4(2818, 5720, 8077) },
            { AOM_CDF4(20716, 26285, 29153) },
            { AOM_CDF4(18494, 25251, 28653) },
            { AOM_CDF4(14384, 22310, 26560) },
            { AOM_CDF4(11169, 18667, 23569) },
            { AOM_CDF4(8587, 15606, 21134) },
            { AOM_CDF4(7356, 13550, 18794) },
            { AOM_CDF4(6277, 11555, 15702) },
            { AOM_CDF4(25791, 30887, 32235) },
            { AOM_CDF4(23193, 29469, 31681) },
            { AOM_CDF4(18326, 26014, 29630) },
            { AOM_CDF4(13377, 22009, 26765) },
            { AOM_CDF4(11252, 19231, 24410) },
            { AOM_CDF4(8855, 16098, 21505) },
            { AOM_CDF4(6382, 12705, 17760) } } },
        { { { AOM_CDF4(13784, 20459, 23807) },
            { AOM_CDF4(11188, 17339, 21036) },
            { AOM_CDF4(7724, 13275, 17419) },
            { AOM_CDF4(5328, 10198, 13773) },
            { AOM_CDF4(4095, 7965, 11303) },
            { AOM_CDF4(3392, 6600, 9642) },
            { AOM_CDF4(1449, 2940, 4455) },
            { AOM_CDF4(19327, 25497, 28310) },
            { AOM_CDF4(17553, 24612, 28071) },
            { AOM_CDF4(13365, 20981, 25167) },
            { AOM_CDF4(10120, 16964, 21644) },
            { AOM_CDF4(7567, 13643, 18303) },
            { AOM_CDF4(5919, 11190, 15483) },
            { AOM_CDF4(2696, 5219, 7571) },
            { AOM_CDF4(24553, 29618, 31290) },
            { AOM_CDF4(22244, 28450, 30726) },
            { AOM_CDF4(17474, 25189, 28681) },
            { AOM_CDF4(13371, 21111, 25560) },
            { AOM_CDF4(10166, 17327, 22080) },
            { AOM_CDF4(8273, 14790, 19458) },
            { AOM_CDF4(4532, 8462, 11904) } },
          { { AOM_CDF4(14949, 20962, 24363) },
            { AOM_CDF4(10935, 17021, 20951) },
            { AOM_CDF4(5643, 11512, 15425) },
            { AOM_CDF4(4777, 8599, 12056) },
            { AOM_CDF4(3766, 6654, 9625) },
            { AOM_CDF4(3491, 6250, 8839) },
            { AOM_CDF4(1185, 2793, 4413) },
            { AOM_CDF4(15352, 21415, 24641) },
            { AOM_CDF4(15073, 22544, 25900) },
            { AOM_CDF4(9968, 16632, 21230) },
            { AOM_CDF4(8613, 14564, 18596) },
            { AOM_CDF4(6692, 11538, 15969) },
            { AOM_CDF4(5970, 10771, 15054) },
            { AOM_CDF4(3974, 7567, 10825) },
            { AOM_CDF4(25316, 30467, 32001) },
            { AOM_CDF4(22258, 28968, 31209) },
            { AOM_CDF4(17604, 25315, 28944) },
            { AOM_CDF4(13228, 21372, 25545) },
            { AOM_CDF4(10802, 18478, 23911) },
            { AOM_CDF4(8865, 15418, 20468) },
            { AOM_CDF4(6715, 12028, 16751) } } },
        { { { AOM_CDF4(9468, 14755, 18209) },
            { AOM_CDF4(7130, 11928, 15157) },
            { AOM_CDF4(6383, 11474, 15200) },
            { AOM_CDF4(4910, 9054, 12928) },
            { AOM_CDF4(4024, 7364, 10586) },
            { AOM_CDF4(2982, 5893, 8964) },
            { AOM_CDF4(1403, 2798, 4342) },
            { AOM_CDF4(12835, 21128, 25312) },
            { AOM_CDF4(10046, 16548, 20629) },
            { AOM_CDF4(8391, 15249, 19805) },
            { AOM_CDF4(7357, 12706, 17214) },
            { AOM_CDF4(6038, 10638, 14573) },
            { AOM_CDF4(4791, 9234, 12520) },
            { AOM_CDF4(2068, 3880, 5623) },
            { AOM_CDF4(16896, 22205, 25412) },
            { AOM_CDF4(20202, 26570, 29067) },
            { AOM_CDF4(16491, 23895, 27565) },
            { AOM_CDF4(12825, 20201, 24683) },
            { AOM_CDF4(9870, 16862, 21517) },
            { AOM_CDF4(8058, 14125, 18888) },
            { AOM_CDF4(3494, 6668, 9441) } },
          { { AOM_CDF4(11445, 17709, 21323) },
            { AOM_CDF4(11711, 17695, 21242) },
            { AOM_CDF4(9138, 15876, 20030) },
            { AOM_CDF4(6215, 12994, 16525) },
            { AOM_CDF4(6809, 13192, 17448) },
            { AOM_CDF4(2185, 5825, 13107) },
            { AOM_CDF4(1317, 3365, 4535) },
            { AOM_CDF4(18950, 29412, 31781) },
            { AOM_CDF4(15327, 24840, 30125) },
            { AOM_CDF4(11864, 18644, 24294) },
            { AOM_CDF4(7399, 12684, 17265) },
            { AOM_CDF4(8192, 14336, 19115) },
            { AOM_CDF4(3390, 7910, 12429) },
            { AOM_CDF4(2587, 4886, 8767) },
            { AOM_CDF4(25732, 31160, 32165) },
            { AOM_CDF4(21798, 29033, 30735) },
            { AOM_CDF4(18482, 26236, 29050) },
            { AOM_CDF4(15709, 23385, 27224) },
            { AOM_CDF4(11133, 19325, 24471) },
            { AOM_CDF4(10412, 18221, 22356) },
            { AOM_CDF4(7025, 12331, 16310) } } },
        { { { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } },
          { { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } } } },
      { { { { AOM_CDF4(18958, 25307, 28119) },
            { AOM_CDF4(14936, 22106, 26063) },
            { AOM_CDF4(8199, 15365, 20003) },
            { AOM_CDF4(5888, 11268, 15877) },
            { AOM_CDF4(4635, 9064, 13128) },
            { AOM_CDF4(3410, 7535, 11086) },
            { AOM_CDF4(2181, 4688, 6970) },
            { AOM_CDF4(22519, 27770, 30318) },
            { AOM_CDF4(20163, 27562, 30336) },
            { AOM_CDF4(15113, 23148, 27188) },
            { AOM_CDF4(10800, 18577, 23888) },
            { AOM_CDF4(8230, 14991, 20344) },
            { AOM_CDF4(6998, 13142, 18159) },
            { AOM_CDF4(4899, 9673, 13712) },
            { AOM_CDF4(24280, 29438, 31568) },
            { AOM_CDF4(21377, 28137, 30752) },
            { AOM_CDF4(16486, 24638, 28774) },
            { AOM_CDF4(12566, 20449, 25387) },
            { AOM_CDF4(9859, 18032, 23031) },
            { AOM_CDF4(8678, 14689, 19545) },
            { AOM_CDF4(5883, 12027, 16289) } },
          { { AOM_CDF4(20198, 27313, 30213) },
            { AOM_CDF4(15491, 23301, 27462) },
            { AOM_CDF4(10862, 18646, 23716) },
            { AOM_CDF4(7564, 13393, 18579) },
            { AOM_CDF4(5902, 11728, 15697) },
            { AOM_CDF4(4223, 9034, 12669) },
            { AOM_CDF4(3928, 6616, 9923) },
            { AOM_CDF4(25378, 30467, 32141) },
            { AOM_CDF4(23465, 29589, 31621) },
            { AOM_CDF4(18438, 26951, 30412) },
            { AOM_CDF4(13588, 21432, 26425) },
            { AOM_CDF4(8826, 15492, 21313) },
            { AOM_CDF4(9128, 14512, 18256) },
            { AOM_CDF4(3542, 11956, 17270) },
            { AOM_CDF4(26119, 31027, 31977) },
            { AOM_CDF4(24753, 30531, 32246) },
            { AOM_CDF4(22098, 28569, 30978) },
            { AOM_CDF4(15099, 26022, 30840) },
            { AOM_CDF4(9175, 15729, 20972) },
            { AOM_CDF4(10923, 21845, 29127) },
            { AOM_CDF4(8192, 16384, 24576) } } },
        { { { AOM_CDF4(19215, 25485, 28552) },
            { AOM_CDF4(14627, 22164, 25968) },
            { AOM_CDF4(8971, 15622, 20509) },
            { AOM_CDF4(6047, 11224, 15688) },
            { AOM_CDF4(4535, 8764, 12458) },
            { AOM_CDF4(3554, 6907, 10398) },
            { AOM_CDF4(1968, 4052, 6172) },
            { AOM_CDF4(19942, 26272, 28703) },
            { AOM_CDF4(18822, 26157, 29303) },
            { AOM_CDF4(14461, 22168, 26504) },
            { AOM_CDF4(10378, 17961, 22909) },
            { AOM_CDF4(8353, 14777, 19740) },
            { AOM_CDF4(6822, 12662, 17145) },
            { AOM_CDF4(4468, 8494, 12287) },
            { AOM_CDF4(24582, 29924, 31542) },
            { AOM_CDF4(22095, 28661, 31008) },
            { AOM_CDF4(17810, 25621, 29157) },
            { AOM_CDF4(13664, 21851, 26466) },
            { AOM_CDF4(10891, 18406, 23387) },
            { AOM_CDF4(9093, 15914, 21002) },
            { AOM_CDF4(6362, 11751, 16334) } },
          { { AOM_CDF4(17933, 25150, 28317) },
            { AOM_CDF4(14239, 21537, 25749) },
            { AOM_CDF4(9140, 15870, 20832) },
            { AOM_CDF4(6192, 12191, 16735) },
            { AOM_CDF4(4743, 9681, 14326) },
            { AOM_CDF4(4203, 7584, 11154) },
            { AOM_CDF4(3263, 6690, 10415) },
            { AOM_CDF4(22007, 27913, 30826) },
            { AOM_CDF4(19865, 27291, 30016) },
            { AOM_CDF4(16073, 24363, 28591) },
            { AOM_CDF4(13377, 21796, 26375) },
            { AOM_CDF4(10220, 17507, 23923) },
            { AOM_CDF4(10246, 15969, 22522) },
            { AOM_CDF4(6622, 14614, 17811) },
            { AOM_CDF4(26801, 30910, 31985) },
            { AOM_CDF4(24783, 30538, 32228) },
            { AOM_CDF4(20026, 27705, 30698) },
            { AOM_CDF4(15832, 23509, 28786) },
            { AOM_CDF4(13184, 22144, 27136) },
            { AOM_CDF4(8777, 20773, 25454) },
            { AOM_CDF4(8048, 15522, 21845) } } },
        { { { AOM_CDF4(14826, 21199, 24951) },
            { AOM_CDF4(12131, 19147, 23377) },
            { AOM_CDF4(8247, 14396, 18825) },
            { AOM_CDF4(5803, 10858, 14609) },
            { AOM_CDF4(4156, 8153, 11623) },
            { AOM_CDF4(3150, 6565, 9494) },
            { AOM_CDF4(1505, 3049, 4665) },
            { AOM_CDF4(19971, 26601, 28818) },
            { AOM_CDF4(17921, 25586, 28803) },
            { AOM_CDF4(13764, 21432, 25786) },
            { AOM_CDF4(10157, 16942, 21682) },
            { AOM_CDF4(8025, 14231, 18547) },
            { AOM_CDF4(6073, 11398, 15693) },
            { AOM_CDF4(3087, 5938, 8576) },
            { AOM_CDF4(25141, 30226, 31771) },
            { AOM_CDF4(22599, 28867, 30997) },
            { AOM_CDF4(17837, 25642, 28870) },
            { AOM_CDF4(13541, 21609, 26045) },
            { AOM_CDF4(10507, 17970, 22646) },
            { AOM_CDF4(8440, 15059, 20001) },
            { AOM_CDF4(5091, 9497, 13357) } },
          { { AOM_CDF4(15295, 21345, 25217) },
            { AOM_CDF4(11361, 18666, 23159) },
            { AOM_CDF4(6123, 13057, 17587) },
            { AOM_CDF4(4920, 8440, 12277) },
            { AOM_CDF4(4855, 7754, 11125) },
            { AOM_CDF4(3002, 6087, 8671) },
            { AOM_CDF4(1853, 4156, 5560) },
            { AOM_CDF4(15486, 22444, 26035) },
            { AOM_CDF4(16111, 22883, 25659) },
            { AOM_CDF4(10738, 18148, 22585) },
            { AOM_CDF4(9762, 16521, 20070) },
            { AOM_CDF4(8385, 13493, 17733) },
            { AOM_CDF4(8561, 13432, 17565) },
            { AOM_CDF4(5936, 10982, 15256) },
            { AOM_CDF4(24333, 31146, 32444) },
            { AOM_CDF4(23054, 29242, 31570) },
            { AOM_CDF4(18369, 25940, 29714) },
            { AOM_CDF4(16230, 23979, 27910) },
            { AOM_CDF4(11685, 20293, 25366) },
            { AOM_CDF4(10596, 18007, 22662) },
            { AOM_CDF4(7742, 14286, 19731) } } },
        { { { AOM_CDF4(7761, 12357, 15729) },
            { AOM_CDF4(6770, 11334, 14888) },
            { AOM_CDF4(6218, 11568, 15496) },
            { AOM_CDF4(4789, 8754, 12604) },
            { AOM_CDF4(3770, 7439, 10801) },
            { AOM_CDF4(3513, 6401, 9817) },
            { AOM_CDF4(1658, 3498, 5189) },
            { AOM_CDF4(14163, 21143, 24925) },
            { AOM_CDF4(12611, 20007, 24098) },
            { AOM_CDF4(10243, 17347, 22345) },
            { AOM_CDF4(7949, 14400, 18977) },
            { AOM_CDF4(6488, 11914, 15984) },
            { AOM_CDF4(4975, 9973, 13827) },
            { AOM_CDF4(2466, 4648, 6733) },
            { AOM_CDF4(19344, 26279, 29443) },
            { AOM_CDF4(19885, 27057, 29851) },
            { AOM_CDF4(16078, 23984, 27842) },
            { AOM_CDF4(12880, 20835, 25181) },
            { AOM_CDF4(9947, 17093, 21828) },
            { AOM_CDF4(8310, 14728, 19427) },
            { AOM_CDF4(3870, 7365, 10275) } },
          { { AOM_CDF4(12988, 19542, 24367) },
            { AOM_CDF4(13238, 20509, 24844) },
            { AOM_CDF4(9452, 14809, 20007) },
            { AOM_CDF4(6242, 11703, 15994) },
            { AOM_CDF4(5085, 10734, 14124) },
            { AOM_CDF4(4096, 5120, 13312) },
            { AOM_CDF4(1872, 5056, 7303) },
            { AOM_CDF4(22640, 28598, 30385) },
            { AOM_CDF4(17217, 25548, 28047) },
            { AOM_CDF4(13336, 20575, 25148) },
            { AOM_CDF4(10128, 14895, 22044) },
            { AOM_CDF4(5201, 15084, 21325) },
            { AOM_CDF4(3511, 8777, 13458) },
            { AOM_CDF4(3492, 6983, 11281) },
            { AOM_CDF4(23526, 29407, 31928) },
            { AOM_CDF4(21666, 28829, 31515) },
            { AOM_CDF4(17375, 23776, 27129) },
            { AOM_CDF4(15867, 22765, 27249) },
            { AOM_CDF4(10835, 17970, 24576) },
            { AOM_CDF4(10639, 20852, 24257) },
            { AOM_CDF4(6642, 11661, 15498) } } },
        { { { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } },
          { { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } } } },
      { { { { AOM_CDF4(20286, 26127, 29060) },
            { AOM_CDF4(16807, 23937, 27782) },
            { AOM_CDF4(10158, 17838, 23197) },
            { AOM_CDF4(7342, 13328, 18192) },
            { AOM_CDF4(4609, 9723, 14397) },
            { AOM_CDF4(3156, 7639, 11414) },
            { AOM_CDF4(2737, 5858, 8646) },
            { AOM_CDF4(22930, 29050, 31064) },
            { AOM_CDF4(21763, 28798, 31345) },
            { AOM_CDF4(17218, 25835, 29624) },
            { AOM_CDF4(12412, 21699, 26547) },
            { AOM_CDF4(9501, 16243, 22420) },
            { AOM_CDF4(7816, 15783, 21720) },
            { AOM_CDF4(6427, 12433, 18017) },
            { AOM_CDF4(26532, 30635, 31729) },
            { AOM_CDF4(24301, 29780, 31501) },
            { AOM_CDF4(19508, 27437, 30953) },
            { AOM_CDF4(15630, 25064, 29307) },
            { AOM_CDF4(10426, 21349, 24990) },
            { AOM_CDF4(13011, 18312, 24576) },
            { AOM_CDF4(11796, 17039, 23593) } },
          { { AOM_CDF4(22981, 29440, 31437) },
            { AOM_CDF4(16902, 25206, 28312) },
            { AOM_CDF4(11077, 18706, 23681) },
            { AOM_CDF4(7022, 14746, 21299) },
            { AOM_CDF4(5350, 11368, 16718) },
            { AOM_CDF4(5699, 9973, 24220) },
            { AOM_CDF4(8192, 10923, 19115) },
            { AOM_CDF4(23177, 28772, 31969) },
            { AOM_CDF4(24242, 28756, 31431) },
            { AOM_CDF4(18404, 26933, 30524) },
            { AOM_CDF4(10486, 15729, 26214) },
            { AOM_CDF4(10923, 21845, 27307) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(20852, 23831, 29789) },
            { AOM_CDF4(20025, 29127, 30948) },
            { AOM_CDF4(16384, 20480, 24576) },
            { AOM_CDF4(6554, 13107, 26214) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } } },
        { { { AOM_CDF4(19557, 25990, 28978) },
            { AOM_CDF4(15096, 22887, 26498) },
            { AOM_CDF4(9157, 16070, 21485) },
            { AOM_CDF4(6681, 12336, 17066) },
            { AOM_CDF4(4914, 9987, 14310) },
            { AOM_CDF4(4175, 7840, 11639) },
            { AOM_CDF4(2624, 5787, 8290) },
            { AOM_CDF4(19680, 26176, 28870) },
            { AOM_CDF4(20770, 28091, 30734) },
            { AOM_CDF4(17304, 25338, 29230) },
            { AOM_CDF4(13267, 21226, 25933) },
            { AOM_CDF4(10659, 18828, 23071) },
            { AOM_CDF4(8931, 15354, 20003) },
            { AOM_CDF4(6811, 11962, 16892) },
            { AOM_CDF4(26088, 30961, 32029) },
            { AOM_CDF4(24413, 29987, 31845) },
            { AOM_CDF4(20646, 28082, 30775) },
            { AOM_CDF4(16772, 24641, 28871) },
            { AOM_CDF4(13881, 21720, 26352) },
            { AOM_CDF4(10238, 18590, 24113) },
            { AOM_CDF4(9064, 15500, 20622) } },
          { { AOM_CDF4(20161, 26862, 30411) },
            { AOM_CDF4(15698, 23673, 28129) },
            { AOM_CDF4(10639, 18831, 23618) },
            { AOM_CDF4(7366, 12955, 16003) },
            { AOM_CDF4(5699, 14247, 19946) },
            { AOM_CDF4(3855, 9638, 15420) },
            { AOM_CDF4(2185, 8738, 15292) },
            { AOM_CDF4(24030, 29127, 31312) },
            { AOM_CDF4(20992, 28416, 30976) },
            { AOM_CDF4(19895, 26917, 29842) },
            { AOM_CDF4(11469, 18022, 24576) },
            { AOM_CDF4(12288, 16384, 20480) },
            { AOM_CDF4(10923, 21845, 27307) },
            { AOM_CDF4(9362, 14043, 18725) },
            { AOM_CDF4(24576, 27853, 31130) },
            { AOM_CDF4(24966, 28867, 31988) },
            { AOM_CDF4(20852, 25321, 28300) },
            { AOM_CDF4(5461, 16384, 21845) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(10923, 16384, 21845) } } },
        { { { AOM_CDF4(18258, 25057, 28291) },
            { AOM_CDF4(13474, 20701, 24865) },
            { AOM_CDF4(7940, 13964, 18778) },
            { AOM_CDF4(5661, 10551, 14282) },
            { AOM_CDF4(4274, 7901, 11333) },
            { AOM_CDF4(2867, 6281, 9967) },
            { AOM_CDF4(2102, 4307, 6422) },
            { AOM_CDF4(18796, 25616, 27695) },
            { AOM_CDF4(19022, 26638, 29265) },
            { AOM_CDF4(14555, 22758, 27249) },
            { AOM_CDF4(10836, 18891, 23387) },
            { AOM_CDF4(8166, 14751, 19808) },
            { AOM_CDF4(7233, 12491, 17550) },
            { AOM_CDF4(5144, 9208, 13012) },
            { AOM_CDF4(25863, 31435, 32344) },
            { AOM_CDF4(23412, 29807, 31705) },
            { AOM_CDF4(19540, 27010, 30118) },
            { AOM_CDF4(15187, 23719, 27936) },
            { AOM_CDF4(12291, 20372, 25089) },
            { AOM_CDF4(10128, 17352, 22379) },
            { AOM_CDF4(7104, 12777, 17403) } },
          { { AOM_CDF4(17820, 23712, 27315) },
            { AOM_CDF4(13195, 20578, 25690) },
            { AOM_CDF4(6746, 13600, 18526) },
            { AOM_CDF4(6687, 11703, 14378) },
            { AOM_CDF4(4161, 7802, 11963) },
            { AOM_CDF4(5595, 9591, 11988) },
            { AOM_CDF4(2016, 3529, 5545) },
            { AOM_CDF4(20632, 24273, 30341) },
            { AOM_CDF4(16846, 22384, 26537) },
            { AOM_CDF4(11663, 19994, 23326) },
            { AOM_CDF4(6827, 13653, 17749) },
            { AOM_CDF4(6554, 13107, 19661) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(18204, 25486, 29127) },
            { AOM_CDF4(23406, 25746, 28087) },
            { AOM_CDF4(25206, 27727, 30247) },
            { AOM_CDF4(18725, 23406, 28087) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } } },
        { { { AOM_CDF4(9404, 15586, 19652) },
            { AOM_CDF4(9286, 15268, 19629) },
            { AOM_CDF4(6652, 12742, 17493) },
            { AOM_CDF4(5620, 10675, 15165) },
            { AOM_CDF4(4601, 8682, 12283) },
            { AOM_CDF4(3679, 7220, 11247) },
            { AOM_CDF4(1515, 3602, 5798) },
            { AOM_CDF4(17538, 26768, 29076) },
            { AOM_CDF4(16683, 24491, 27823) },
            { AOM_CDF4(13944, 21392, 25937) },
            { AOM_CDF4(10852, 17235, 22257) },
            { AOM_CDF4(6780, 12562, 17747) },
            { AOM_CDF4(6087, 10043, 14710) },
            { AOM_CDF4(4058, 8290, 11432) },
            { AOM_CDF4(23954, 29761, 31524) },
            { AOM_CDF4(21992, 28343, 30572) },
            { AOM_CDF4(18097, 25787, 29183) },
            { AOM_CDF4(14507, 22607, 26994) },
            { AOM_CDF4(10912, 19313, 23594) },
            { AOM_CDF4(9279, 15164, 20096) },
            { AOM_CDF4(6245, 10534, 13894) } },
          { { AOM_CDF4(16326, 23042, 26979) },
            { AOM_CDF4(16384, 23749, 27657) },
            { AOM_CDF4(9947, 18139, 23991) },
            { AOM_CDF4(6242, 15604, 21845) },
            { AOM_CDF4(4096, 9557, 15019) },
            { AOM_CDF4(5120, 7168, 13312) },
            { AOM_CDF4(3449, 7761, 12072) },
            { AOM_CDF4(21065, 25746, 30427) },
            { AOM_CDF4(15381, 22737, 26081) },
            { AOM_CDF4(12072, 22420, 28169) },
            { AOM_CDF4(8495, 15777, 19418) },
            { AOM_CDF4(7282, 14564, 21845) },
            { AOM_CDF4(5174, 10348, 13797) },
            { AOM_CDF4(4274, 7123, 11398) },
            { AOM_CDF4(18725, 23406, 28087) },
            { AOM_CDF4(18432, 21504, 25600) },
            { AOM_CDF4(10426, 25321, 31279) },
            { AOM_CDF4(3277, 16384, 19661) },
            { AOM_CDF4(9362, 18725, 23406) },
            { AOM_CDF4(10923, 16384, 21845) },
            { AOM_CDF4(10923, 18204, 21845) } } },
        { { { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } },
          { { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) }, { AOM_CDF4(8192, 16384, 24576) },
            { AOM_CDF4(8192, 16384, 24576) } } } },
    };
